/*
 * ctrl_impl.c
 *
 *  Created on: 15.02.2016
 *      Author: AndreR
 */

#include "ctrl_impl.h"

#include <main/ctrl.h>
#include <main/skills.h>
#include "hal/buzzer.h"
#include "hal/kicker.h"
#include "hal/motors.h"

#include "util/console.h"
#include "util/sys_time.h"
#include "util/init_hal.h"
#include "util/intercom.h"
#include "util/arm_mat_util_f32.h"
#include "util/intercom_reliable.h"
#include "util/log_file.h"
#include "main/network.h"

#include "constants.h"
#include "commands.h"
#include "power.h"
#include "spi6.h"

CtrlImpl ctrlImpl;

BotParameters botParams = {
	.driveTrain = {
		.motor2WheelRatio = (18.0f/60.0f),
		.wheel2MotorRatio = (60.0f/18.0f),
		.motor = {
			.Km = 0.0251f,
			.Kn = (380.0f*2.0f*PI/60.0f),
			.Ke = (60.0f/(380.0f*2.0f*PI)),
			.Kf = 0.94e-4f,
			.R = 0.464f,
			.L = 322e-6f,
		}
	},
	.physical = { 0.0285f, 30.0f, 45.0f, 0.083f, 2.3f },
};

static const float out2MotX[129];
static const float out2MotY[129];
static const float out2MotW[129];
static const float mot2InX[129];
static const float mot2InY[129];

void CtrlImplInit()
{
}

void CtrlImplKickerFire(uint8_t mode, uint8_t device, float speed)
{
	switch(mode)
	{
	case IC_KD_FIRE_MODE_FORCE:
		KickerFireSpeed(speed, device);
		break;
	case IC_KD_FIRE_MODE_ARM:
		KickerArm(speed, device);
		break;
	case IC_KD_FIRE_MODE_DISARM:
		KickerDisarm();
		break;
	}
}

void CtrlImplWriteLogMsg(void* pMsg)
{
	LogFileWrite(pMsg);
}

void CtrlImplUpdateInput(CtrlInput* pInput, LogKickerDribbler* pKD)
{
	static uint32_t lastVisionTimestamp;

	// motors
	pInput->enc.time = motors.drive[0].velocityTimestamp;
	pInput->motorVol.time = motors.drive[0].outVoltageTimestamp;

	for(uint8_t m = 0; m < 4; m++)
	{
		pInput->enc.vel[m] = motors.drive[m].velocity;
		pInput->motorVol.vol[m] = motors.drive[m].outVoltage;
	}

	// power
	pInput->power.voltage = power.vBat;
	pInput->power.current = power.iCur;
	pInput->power.batEmpty = power.batEmpty;

	// IMU
	pInput->acc.linAcc[0] = spi6.imu.acc[1];
	pInput->acc.linAcc[1] = -spi6.imu.acc[0];
	pInput->acc.linAcc[2] = spi6.imu.acc[2];
	pInput->acc.time = SysTimeUSec();
	pInput->acc.updated = 1;

	pInput->gyr.rotVel[0] = spi6.imu.gyr[1];
	pInput->gyr.rotVel[1] = -spi6.imu.gyr[0];
	pInput->gyr.rotVel[2] = spi6.imu.gyr[2];
	pInput->gyr.time = SysTimeUSec();
	pInput->gyr.updated = 1;

	// vision
	if(lastVisionTimestamp != network.lastVisionPosTimestamp)
	{
		pInput->vision.pos[0] = network.lastVisionPos[0]*0.001f;
		pInput->vision.pos[1] = network.lastVisionPos[1]*0.001f;
		pInput->vision.pos[2] = network.lastVisionPos[2]*0.001f;
		pInput->vision.time = network.lastVisionPosTimestamp;
		pInput->vision.delay = network.lastVisionDelay;
		pInput->vision.camId = network.lastVisionCamId;
		pInput->vision.updated = 1;

		lastVisionTimestamp = network.lastVisionPosTimestamp;
	}

	// KD
	pKD->kickerLevel = kicker.vCap;
	pKD->kickerChgCurrent = kicker.iChg;
	pKD->kickCounter = kicker.kickCounter;

	pKD->dribblerSpeed = (int16_t)(motors.dribbler.velocity*MOTORS_DRIBBLER_MOTOR_RADS_TO_BAR_RPM);
	pKD->dribblerTemp = 0;	// TODO: implement for new motor
	pKD->dribblerOverheated = 0;

	pKD->irLevel[0] = (uint16_t)(kicker.vIrOff*1000.0f);
	pKD->irLevel[1] = (uint16_t)(kicker.vIrOn*1000.0f);
	pKD->barrierInterrupted = kicker.interrupted;

	pKD->header.timestamp = SysTimeUSec();
}

void CtrlImplApplyOutput(const CtrlOutput* pOutput)
{
	if(motors.acceptCtrlOutput == 0)
		return;

	motors.driveMode = pOutput->ctrlMode;

	for(uint8_t i = 0; i < 4; i++)
	{
		motors.drive[i].setVelocity = pOutput->motorVel[i];
		motors.drive[i].setVoltage = pOutput->motorVoltage[i];
	}
}

void CtrlImplSendNetworkPacket(PacketHeader* pHeader, uint8_t* pData, uint16_t dataLength)
{
	NetworkSendPacket(pHeader, pData, dataLength);
}

void CtrlImplBuzzerPlay(uint16_t seqId)
{
	switch(seqId)
	{
	case CTRL_BUZZ_UP20: BuzzerPlay(&buzzSeqUp20); break;
	case CTRL_BUZZ_BEEP_FAST: BuzzerPlay(&buzzSeqBeepFast); break;
	case CTRL_BUZZ_UP50: BuzzerPlay(&buzzSeqUp50); break;
	case CTRL_BUZZ_DOWN50: BuzzerPlay(&buzzSeqDown50); break;
	case CTRL_BUZZ_UPDOWN20: BuzzerPlay(&buzzSeqUpDown20); break;
	}
}

void CtrlImplCorrectOutXYW(float* pLocalVel)
{
	const int32_t numSupportPoints = sizeof(out2MotX)/sizeof(out2MotX[0]);

	float x = pLocalVel[0];
	float y = pLocalVel[1];

	float mod = 1;
	if(x < 0)
		mod = -1;

	float speed = sqrtf(x*x+y*y);

	if(speed < 0.001f)
		return;

	float dir = arm_atan2_f32(y, x*mod);

	float exactPos = (dir+PI/2)/PI*(numSupportPoints-1); // 0.0 - 1.0 => 0 - 129

	int32_t leftIndex = (int32_t)exactPos;
	int32_t rightIndex = (int32_t)(exactPos+1);

	float rightUsage = exactPos - leftIndex;
	float leftUsage = 1.0f - rightUsage;

	if(leftIndex < 0)
	{
		leftIndex = 0;
		rightIndex = 1;
	}

	if(rightIndex >= numSupportPoints)
	{
		rightIndex = numSupportPoints-1;
		leftIndex = numSupportPoints-2;
	}

//	ConsolePrint("L: %d, R: %d, Lu: %f, Ru: %f\r\n", leftIndex, rightIndex, leftUsage, rightUsage);

	float corr[3];
	corr[0] = leftUsage*out2MotX[leftIndex] + rightUsage*out2MotX[rightIndex];
	corr[1] = leftUsage*out2MotY[leftIndex] + rightUsage*out2MotY[rightIndex];
	corr[2] = leftUsage*out2MotW[leftIndex] + rightUsage*out2MotW[rightIndex];

	pLocalVel[0] = speed*corr[0]*mod;
	pLocalVel[1] = speed*corr[1];
//	pLocalVel[2] = pLocalVel[2] + speed*corr[2]*mod;
}

void CtrlImplCorrectInXYW(float* pLocalVel)
{
	const int32_t numSupportPoints = sizeof(out2MotX)/sizeof(out2MotX[0]);

	float x = pLocalVel[0];
	float y = pLocalVel[1];

	float mod = 1;
	if(x < 0)
		mod = -1;

	float speed = sqrtf(x*x+y*y);

	if(speed < 0.001f)
		return;

	float dir = arm_atan2_f32(y, x*mod);

	float exactPos = (dir+PI/2)/PI*(numSupportPoints-1); // 0.0 - 1.0 => 0 - 129

	int32_t leftIndex = (int32_t)exactPos;
	int32_t rightIndex = (int32_t)(exactPos+1);

	float rightUsage = exactPos - leftIndex;
	float leftUsage = 1.0f - rightUsage;

	if(leftIndex < 0)
	{
		leftIndex = 0;
		rightIndex = 1;
	}

	if(rightIndex >= numSupportPoints)
	{
		rightIndex = numSupportPoints-1;
		leftIndex = numSupportPoints-2;
	}

	float corr[3];
	corr[0] = leftUsage*mot2InX[leftIndex] + rightUsage*mot2InX[rightIndex];
	corr[1] = leftUsage*mot2InY[leftIndex] + rightUsage*mot2InY[rightIndex];
	corr[2] = leftUsage*out2MotW[leftIndex] + rightUsage*out2MotW[rightIndex];

	pLocalVel[0] = speed*corr[0]*mod;
	pLocalVel[1] = speed*corr[1];
//	pLocalVel[2] = pLocalVel[2] - speed*corr[2]*mod;
}


static const float out2MotX[129] = {
0.00000000, 0.02067798, 0.04134117, 0.06197480, 0.08256413, 0.10352448, 0.12441111, 0.14520934,
0.16590460, 0.18807348, 0.21011038, 0.23200093, 0.25373081, 0.27828322, 0.30264668, 0.32680721,
0.35075095, 0.37832748, 0.40565991, 0.43273480, 0.45953885, 0.48869828, 0.51756080, 0.54611365,
0.57434424, 0.60291755, 0.63114402, 0.65901167, 0.68650872, 0.71454397, 0.74218578, 0.76942306,
0.79624500, 0.82380029, 0.85091931, 0.87759202, 0.90380862, 0.92811780, 0.95195211, 0.97530260,
0.99816061, 1.01678407, 1.03489862, 1.05249650, 1.06957028, 1.08253382, 1.09495931, 1.10684028,
1.11817055, 1.12655490, 1.13437725, 1.14163244, 1.14831568, 1.15320416, 1.15751211, 1.16123574,
1.16437164, 1.16648009, 1.16799504, 1.16891414, 1.16923542, 1.16889652, 1.16795690, 1.16641563,
1.16427219, 1.16141357, 1.15795278, 1.15389034, 1.14922718, 1.14356701, 1.13730902, 1.13045518,
1.12300784, 1.11434443, 1.10509329, 1.09525784, 1.08484185, 1.07320354, 1.06099327, 1.04821584,
1.03487642, 1.02082392, 1.00622072, 0.99107300, 0.97538723, 0.95980291, 0.94369448, 0.92706938,
0.90993538, 0.89340748, 0.87638711, 0.85888290, 0.84090383, 0.82333801, 0.80531607, 0.78684781,
0.76794330, 0.74882929, 0.72929993, 0.70936604, 0.68903871, 0.66796598, 0.64652266, 0.62472051,
0.60257152, 0.57986373, 0.55683368, 0.53349396, 0.50985733, 0.48640133, 0.46267450, 0.43869013,
0.41446166, 0.39079222, 0.36690599, 0.34281684, 0.31853873, 0.29421455, 0.26972970, 0.24509850,
0.22033531, 0.19419456, 0.16795082, 0.14161870, 0.11521284, 0.08646134, 0.05766552, 0.02884015,
0.00000000,  };

static const float out2MotY[129] = {
-1.00791016, -1.00792468, -1.00733702, -1.00614772, -1.00435769, -1.00229751, -0.99963950, -0.99638565,
-0.99253830, -0.98781141, -0.98249679, -0.97659786, -0.97011838, -0.96235648, -0.95402263, -0.94512161,
-0.93565860, -0.92452213, -0.91283497, -0.90060329, -0.88783355, -0.87330615, -0.85825463, -0.84268645,
-0.82660936, -0.80910559, -0.79110933, -0.77262925, -0.75367430, -0.73356138, -0.71299235, -0.69197700,
-0.67052540, -0.64846538, -0.62599001, -0.60311012, -0.57983678, -0.55692134, -0.53363531, -0.50999046,
-0.48599876, -0.46302933, -0.43973763, -0.41613626, -0.39223798, -0.36944975, -0.34639069, -0.32307408,
-0.29951337, -0.27693544, -0.25414071, -0.23114306, -0.20795645, -0.18555740, -0.16299770, -0.14029164,
-0.11745359, -0.09521889, -0.07288121, -0.05045515, -0.02795535, -0.00592238, 0.01615490, 0.03826174,
0.06038336, 0.08217960, 0.10396105, 0.12571295, 0.14742054, 0.16896758, 0.19044089, 0.21182581,
0.23310775, 0.25433575, 0.27543178, 0.29638144, 0.31717044, 0.33805281, 0.35874622, 0.37923671,
0.39951041, 0.41993806, 0.44012161, 0.46004761, 0.47970279, 0.49949414, 0.51898859, 0.53817336,
0.55703587, 0.57589000, 0.59439729, 0.61254576, 0.63032363, 0.64787989, 0.66504271, 0.68180100,
0.69814394, 0.71413877, 0.72969735, 0.74480960, 0.75946574, 0.77381953, 0.78769844, 0.80109353,
0.81399614, 0.82685846, 0.83921187, 0.85104861, 0.86236125, 0.87412432, 0.88534935, 0.89602984,
0.90615965, 0.91694439, 0.92716712, 0.93682270, 0.94590632, 0.95526777, 0.96404867, 0.97224526,
0.97985413, 0.98626577, 0.99208391, 0.99730621, 1.00193068, 1.00144660, 1.00036179, 0.99867535,
0.99638672,  };

static const float out2MotW[129] = {
-0.00195592, 0.00648364, 0.01459833, 0.02222444, 0.02922257, 0.03548152, 0.04092096, 0.04549285,
0.04918147, 0.05200213, 0.05399872, 0.05523999, 0.05581500, 0.05582778, 0.05539147, 0.05462227,
0.05363335, 0.05252913, 0.05140002, 0.05031803, 0.04933324, 0.04847143, 0.04773297, 0.04709282,
0.04650192, 0.04588967, 0.04516747, 0.04423318, 0.04297629, 0.04128344, 0.03904430, 0.03615724,
0.03253474, 0.02810827, 0.02283233, 0.01668748, 0.00968235, 0.00185438, -0.00673075, -0.01598083,
-0.02578005, -0.03599301, -0.04646958, -0.05705044, -0.06757307, -0.07787796, -0.08781465, -0.09724755,
-0.10606103, -0.11416381, -0.12149226, -0.12801257, -0.13372169, -0.13864691, -0.14284420, -0.14639532,
-0.14940381, -0.15199002, -0.15428549, -0.15642675, -0.15854895, -0.16077946, -0.16323186, -0.16600044,
-0.16915549, -0.17273964, -0.17676529, -0.18121332, -0.18603317, -0.19114413, -0.19643804, -0.20178299,
-0.20702813, -0.21200923, -0.21655485, -0.22049278, -0.22365659, -0.22589191, -0.22706234, -0.22705452,
-0.22578243, -0.22319059, -0.21925600, -0.21398895, -0.20743251, -0.19966076, -0.19077600, -0.18090484,
-0.17019359, -0.15880294, -0.14690238, -0.13466437, -0.12225872, -0.10984732, -0.09757932, -0.08558726,
-0.07398392, -0.06286032, -0.05228468, -0.04230258, -0.03293801, -0.02419555, -0.01606319, -0.00851592,
-0.00151975, 0.00496406, 0.01097473, 0.01654789, 0.02171193, 0.02648500, 0.03087273, 0.03486681,
0.03844466, 0.04156992, 0.04419408, 0.04625892, 0.04769976, 0.04844940, 0.04844244, 0.04761996,
0.04593408, 0.04335241, 0.03986202, 0.03547276, 0.03021975, 0.02416486, 0.01739715, 0.01003207,
0.00220960,  };

static const float mot2InX[129] = {
0.00001875, 0.02908157, 0.05823996, 0.08736487, 0.11631941, 0.14496386, 0.17316070, 0.20077973,
0.22770276, 0.25382771, 0.27907198, 0.30337478, 0.32669849, 0.34902878, 0.37037372, 0.39076175,
0.41023881, 0.42886466, 0.44670872, 0.46384559, 0.48035056, 0.49629533, 0.51174421, 0.52675095,
0.54135640, 0.55558719, 0.56945529, 0.58295866, 0.59608278, 0.60880305, 0.62108778, 0.63290168,
0.64420956, 0.65498001, 0.66518884, 0.67482205, 0.68387815, 0.69236965, 0.70032371, 0.70778178,
0.71479835, 0.72143883, 0.72777668, 0.73388999, 0.73985769, 0.74575556, 0.75165241, 0.75760647,
0.76366240, 0.76984884, 0.77617697, 0.78263981, 0.78921260, 0.79585398, 0.80250813, 0.80910752,
0.81557630, 0.82183402, 0.82779945, 0.83339437, 0.83854706, 0.84319526, 0.84728853, 0.85078983,
0.85367620, 0.85593864, 0.85758116, 0.85861895, 0.85907612, 0.85898280, 0.85837204, 0.85727665,
0.85572609, 0.85374381, 0.85134497, 0.84853482, 0.84530790, 0.84164787, 0.83752828, 0.83291392,
0.82776295, 0.82202940, 0.81566607, 0.80862757, 0.80087330, 0.79237014, 0.78309477, 0.77303542,
0.76219287, 0.75058076, 0.73822506, 0.72516287, 0.71144048, 0.69711090, 0.68223103, 0.66685853,
0.65104879, 0.63485202, 0.61831067, 0.60145754, 0.58431438, 0.56689143, 0.54918764, 0.53119178,
0.51288424, 0.49423955, 0.47522926, 0.45582524, 0.43600302, 0.41574505, 0.39504353, 0.37390285,
0.35234121, 0.33039143, 0.30810091, 0.28553059, 0.26275295, 0.23984931, 0.21690633, 0.19401199,
0.17125138, 0.14870230, 0.12643110, 0.10448897, 0.08290886, 0.06170318, 0.04086258, 0.02035574,
0.00013027,  };

static const float mot2InY[129] = {
-0.99218989, -0.99180281, -0.99050035, -0.98838633, -0.98555061, -0.98207006, -0.97800955, -0.97342292,
-0.96835415, -0.96283837, -0.95690296, -0.95056861, -0.94385028, -0.93675823, -0.92929884, -0.92147547,
-0.91328918, -0.90473938, -0.89582439, -0.88654191, -0.87688940, -0.86686440, -0.85646468, -0.84568845,
-0.83453438, -0.82300161, -0.81108972, -0.79879862, -0.78612844, -0.77307939, -0.75965156, -0.74584481,
-0.73165857, -0.71709170, -0.70214238, -0.68680802, -0.67108517, -0.65496948, -0.63845578, -0.62153806,
-0.60420961, -0.58646316, -0.56829101, -0.54968532, -0.53063830, -0.51114251, -0.49119111, -0.47077824,
-0.44989927, -0.42855114, -0.40673266, -0.38444479, -0.36169091, -0.33847707, -0.31481214, -0.29070801,
-0.26617967, -0.24124529, -0.21592617, -0.19024674, -0.16423440, -0.13791940, -0.11133455, -0.08451504,
-0.05749803, -0.03032234, -0.00302805, 0.02434392, 0.05175226, 0.07915568, 0.10651343, 0.13378566,
0.16093390, 0.18792143, 0.21471363, 0.24127832, 0.26758600, 0.29361009, 0.31932703, 0.34471637,
0.36976078, 0.39444600, 0.41876065, 0.44269608, 0.46624603, 0.48940637, 0.51217465, 0.53454968,
0.55653106, 0.57811869, 0.59931222, 0.62011059, 0.64051147, 0.66051081, 0.68010242, 0.69927755,
0.71802460, 0.73632888, 0.75417246, 0.77153415, 0.78838960, 0.80471144, 0.82046969, 0.83563216,
0.85016510, 0.86403392, 0.87720407, 0.88964202, 0.90131640, 0.91219922, 0.92226720, 0.93150315,
0.93989749, 0.94744969, 0.95416988, 0.96008032, 0.96521691, 0.96963068, 0.97338909, 0.97657735,
0.97929954, 0.98167952, 0.98386177, 0.98601196, 0.98831721, 0.99098627, 0.99424926, 0.99835719,
1.00358124,  };
