/*
 * test.c
 *
 *  Created on: 27.06.2016
 *      Author: AndreR
 */

#include "test.h"
#include "hal/motors.h"
#include "hal/led.h"
#include "power.h"
#include "main/network.h"
#include "arm_math.h"
#include "util/arm_mat_util_f32.h"
#include "gui/test_drive.h"
#include "util/console.h"

TestData test;

void TestInit()
{
	chMBObjectInit(&test.eventQueue, test.eventQueueData, TEST_EVENT_QUEUE_SIZE);

	test.pDriveProg = &TestDriveProgress;
}

void TestTask(void* params)
{
	(void)params;

	msg_t event;

	chRegSetThreadName("Test");

	while(1)
	{
		if(chMBFetch(&test.eventQueue, &event, TIME_INFINITE) != MSG_OK)
			continue;

		uint8_t testId = event & 0xFF;
		uint8_t motorId = event >> 8;

		switch(testId)
		{
			case IC_TEST_MOT_DYN:
			{
				TestMotorDynamicsResult result;
				TestMotorDynamics(motorId, &result);

				TestDriveDynamicResult(&result);
			}
			break;
		}
	}
}

void TestScheduleMotorTest(uint8_t testId, uint8_t motorId)
{
	msg_t event = motorId;
	event <<= 8;
	event |= testId;

	chMBPost(&test.eventQueue, event, TIME_IMMEDIATE);
}

#define EST_SIZE 12
void TestMotorDynamics(uint8_t motorId, TestMotorDynamicsResult* pResult)
{
	EMAFilter current;
	EMAFilter speed;

	arm_matrix_instance_f32 A = {EST_SIZE, 2, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 Ai = {2, EST_SIZE, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 b = {EST_SIZE, 1, (float[EST_SIZE*1]){}};
	arm_matrix_instance_f32 x = {2, 1, (float[2]){}};

	TestProgress prog;
	prog.testId = IC_TEST_MOT_DYN;
	prog.motorId = motorId;
	prog.percentComplete = 0;

	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

	pResult->motorId = motorId;

	motors.acceptCtrlOutput = 0;
	motors.dribbler.enableCtrl = 0;
	LEDBlank(1);
	network.botDisabled = 1;
	MotorsDribblerEnable(1);
	chThdSleepMilliseconds(5);

	arm_mat_zero_f32(&A);
	arm_mat_zero_f32(&Ai);
	arm_mat_zero_f32(&b);
	arm_mat_zero_f32(&x);

	EMAFilterInit(&current, 0.9f);
	EMAFilterInit(&speed, 0.95f);

	motors.driveMode = MOTORS_DRIVE_MODE_VOLTAGE;
	for(uint8_t i = 0; i < 4; i++)
		motors.drive[i].setVoltage = 0;

	MotorsDribblerSetVoltage(0);

	for(uint16_t j = 0; j < 1000; j++)
	{
		EMAFilterUpdate(&current, power.iCur);
		chThdSleepMilliseconds(1);
	}

	float mainOffset = current.value;

	prog.percentComplete = 1/18.0f*100.0f;
	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

//	ConsolePrint("Main current offset: %d\r\n", (int32_t)(mainOffset*1000.0f));

	for(int16_t i = 0; i <= EST_SIZE; i++)
	{
		float volt = i-EST_SIZE/2;

		if(i == EST_SIZE/2)
			continue;

		if(motorId > 3)
			MotorsDribblerSetVoltage(fabsf(volt));
		else
			motors.drive[motorId].setVoltage = volt;

		chThdSleepMilliseconds(200);

		for(uint16_t j = 0; j < 800; j++)
		{
			EMAFilterUpdate(&current, power.iCur-mainOffset);

			if(motorId > 3)
				EMAFilterUpdate(&speed, motors.dribbler.velocity);
			else
				EMAFilterUpdate(&speed, motors.drive[motorId].velocity);

			chThdSleepMilliseconds(1);
		}

//		float mainCur = pDrive->bat*mainCurrent.value/volt;

		if(volt < 0)
			MAT_ELEMENT(A, i-1, 0) = -current.value;
		else
			MAT_ELEMENT(A, i-1, 0) = current.value;

		if(volt < 0 && motorId > 3)
			MAT_ELEMENT(A, i-1, 1) = -speed.value;
		else
			MAT_ELEMENT(A, i-1, 1) = speed.value;
		MAT_ELEMENT(b, i-1, 0) = volt;

		prog.percentComplete = ((float)(i+2))/18.0f*100.0f;
		if(test.pDriveProg)
			(*test.pDriveProg)(&prog);

//		ConsolePrint("U: %dmV, I: %dmA, v: %dmrad/s\r\n",
//				(int32_t)(volt*1000.0f), (int32_t)(current.value*1000.0f), (int32_t)(speed.value*1000.0f));
	}

	if(motorId > 3)
		MotorsDribblerSetVoltage(0);
	else
		motors.drive[motorId].setVoltage = 0;

	arm_mat_pinv(&A, &Ai);
	arm_mat_mult_f32(&Ai, &b, &x);

	float R = MAT_ELEMENT(x, 0, 0);
	float k = MAT_ELEMENT(x, 1, 0);

	pResult->R = R;
	pResult->k = k;
	pResult->cur = current.value;
	pResult->speed = speed.value;

	prog.percentComplete = 100;
	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

	motors.acceptCtrlOutput = 1;
	motors.dribbler.enableCtrl = 1;
	MotorsDribblerEnable(0);
	LEDBlank(0);
	network.botDisabled = 0;

	chThdSleepMilliseconds(1000);
}
