#include "hal/led.h"
#include "hal/usart2.h"
#include "hal/exti.h"
#include "hal/fsmc.h"
#include "hal/sram.h"
#include "hal/audio.h"
#include "hal/touch.h"
#include "hal/ov7725.h"

#include "cli.h"
#include "presenter.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "usb/usb.h"
#include "usb/usb_hcd.h"
#include "usb/usb_msc.h"
#include "spi1.h"
#include "test.h"
#include "cpu_load.h"

#include "util/boot.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/crc.h"
#include "util/network_print.h"
#include "util/sys_time.h"
#include "util/intercom.h"
#include "util/intercom_reliable.h"
#include "util/prog_client.h"
#include "util/config.h"
#include "util/log_file.h"

#include "ch.h"

#include <stdlib.h>

static THD_WORKING_AREA(waUSART2, 1024);
static THD_WORKING_AREA(waIntercom, 4096);
#ifdef ENV_RUN
static THD_WORKING_AREA(waCLI, 6000);
static THD_WORKING_AREA(waLogFile, 1024);
static THD_WORKING_AREA(_ccm_waUSB, 4096);
static THD_WORKING_AREA(_ccm_waUSBHCD, 4096);
static THD_WORKING_AREA(_ccm_waPresenter, 4096);
static THD_WORKING_AREA(waConfig, 512);
static THD_WORKING_AREA(waTest, 2048);
static THD_WORKING_AREA(waSPI1, 4096);
#endif

#define LOG_FILE_BUFFER_SIZE	32768
#define LOG_FILE_BUFFER_CHUNKS	4
static __attribute__ ((section (".extram"))) uint8_t logFileWriteBuf[FIFO_BLOCK_DATA(LOG_FILE_BUFFER_CHUNKS, LOG_FILE_BUFFER_SIZE)];

static void setExt1High()
{
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.pupd = GPIO_PUPD_NONE;

	GPIOInit(GPIOB, GPIO_PIN_14, &gpioInit);

	GPIOSet(GPIOB, GPIO_PIN_14);
}

static void setExt2High()
{
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.pupd = GPIO_PUPD_NONE;

	GPIOInit(GPIOB, GPIO_PIN_0, &gpioInit);

	GPIOSet(GPIOB, GPIO_PIN_0);
}

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	// Enable important clocks
	RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN | RCC_AHB1ENR_DMA2EN | RCC_AHB1ENR_GPIOAEN
			| RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN
			| RCC_AHB1ENR_GPIOEEN | RCC_AHB1ENR_GPIOFEN | RCC_AHB1ENR_GPIOGEN
			| RCC_AHB1ENR_CRCEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	// Enable compensation cell to reduce noise on power lines with 50MHz/100MHz I/Os
	SYSCFG->CMPCR = SYSCFG_CMPCR_CMP_PD;

	// Enable usage fault, bus fault, and mem fault
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

	setExt1High();	// signal this as early as possible to let main know there is an ext. board

	SystemClockInitData clkInit;
	clkInit.pll.M = 12;
	clkInit.pll.N = 336;
	clkInit.pll.P = 2;
	clkInit.pll.Q = 7;
	clkInit.HSEBypass = 0;
	clkInit.APB1Div = 4;	// = 42MHz
	clkInit.APB2Div = 4;	// = 42MHz
	clkInit.RTCDiv = 12;	// = 1MHz
	clkInit.flashLatency = 5;
	clkInit.sysTickFreq = CH_CFG_ST_FREQUENCY;
	SystemClockInit(&clkInit, 168000000UL);

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	chSysInit();

	// Configure and start Independent watchdog
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
//	IWDG_SetPrescaler(IWDG_Prescaler_8);
//	IWDG_SetReload(2500);	// approx. 500ms
//	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);

	// -- HAL
	LEDInit();
	CRCInit();
	FlashFSInit(0x08060000, 0x20000);	// 128k
	ConfigInit(&ConfigOutputIntercom, INTERCOM_ADDRESS_MEDIA);
	ExtiInit();
	USART2Init();
	AudioInit();
	Intercom1Init();
	IntercomReliableInit();
	SysTimeInitCorrection(10);
	ProgClientInit();
	ConfigIntercomInit();

#ifdef ENV_RUN
	CLIInit();
	NetworkPrintInitIntercom();
	SPI1Init();
	TestInit();
#endif

	chThdCreateStatic(waUSART2, sizeof(waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(waIntercom, sizeof(waIntercom), NORMALPRIO-1, IntercomTask, 0);

#ifdef ENV_RUN
	chThdCreateStatic(waCLI, sizeof(waCLI), NORMALPRIO-2, CLITask, 0);
	chThdCreateStatic(waConfig, sizeof(waConfig), NORMALPRIO-8, ConfigTask, 0);
	chThdCreateStatic(waTest, sizeof(waTest), NORMALPRIO-20, TestTask, 0);
#endif

	setExt2High();	// signal that intercom is ready

	chThdSleepMilliseconds(1000);

	// Second stage init
	FSMCInit();
	SRAMInit();

#ifdef ENV_RUN
	LogFileInit(logFileWriteBuf, LOG_FILE_BUFFER_CHUNKS, LOG_FILE_BUFFER_SIZE);
	USBInit();
	USBHCDInit();
	USBMSCInit();

	TouchInit();
	PresenterInit();
//	OV7725Init();

	chThdCreateStatic(waLogFile, sizeof(waLogFile), NORMALPRIO-3, LogFileTask, 0);
	chThdCreateStatic(_ccm_waUSB, sizeof(_ccm_waUSB), NORMALPRIO-5, USBTask, 0);
	chThdCreateStatic(_ccm_waUSBHCD, sizeof(_ccm_waUSBHCD), NORMALPRIO-6, USBHCDTask, 0);
	chThdCreateStatic(_ccm_waPresenter, sizeof(_ccm_waPresenter), NORMALPRIO-1, PresenterTask, 0);
	chThdCreateStatic(waSPI1, sizeof(waSPI1), NORMALPRIO-10, SPI1Task, 0);
#endif

#ifdef ENV_BOOT
//	ConsolePrint("--- Bootloader (MEDIA) ---\r\n");
#else
	ConsolePrint("--- Media ---\r\n");
#endif

	chThdSetPriority(HIGHPRIO-1);

	while(1)
	{
		chThdSleepMilliseconds(10);

#ifdef ENV_RUN
		uint32_t time = SysTimeUSec()%1000000;

		if(time > 600000 && time < 800000)
			LEDMustOn(LED_GREEN);
		else
			LEDMustOff(LED_GREEN);
#else
		uint32_t time = SysTimeUSec()%500000;

		if(time > 300000 && time < 400000)
			LEDMustOn(LED_RED);
		else
			LEDMustOff(LED_RED);
#endif

		CPULoadUpdateUsage();
	}

	return 0;
}
