/*
 * presenter.c
 *
 *  Created on: 25.03.2014
 *      Author: AndreR
 */

#include "presenter.h"

#include "util/console.h"
#include "util/intercom.h"
#include "hal/touch.h"
#include "usb/usb_hcd.h"
#include "usb/usb_msc.h"
#include "log_msgs.h"
#include "util/log_file.h"
#include "util/network_print.h"

#include "gui/bot_cover.h"
#include "gui/main_status.h"
#include "gui/top_bar.h"
#include "gui/menu.h"
#include "gui/wifi.h"
#include "gui/select_botid.h"
#include "gui/ids.h"
#include "gui/logging.h"
#include "gui/features.h"
#include "gui/optical.h"
#include "gui/test_drive.h"
#include "gui/test_dribbler.h"

#include "gfx.h"
#include "hal/tft.h"

#include <string.h>
#include <stdio.h>

Presenter presenter;

#define GUI_HEAP_SIZE (24*1024)
memory_heap_t guiHeap;
uint8_t guiHeapBuffer[GUI_HEAP_SIZE] __attribute__((aligned(sizeof(stkalign_t))));

typedef int16_t(*EventHandlerFunc)(GEvent*);

static struct _handles
{
	GHandle hTopBar;
	GHandle hMenu;

	GHandle hWindows[9];
} handles;

static void presenterCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case IC_PRESENTER_CONFIG:
		{
			MainStatusWifiUpdate((ConfigNetwork*)pData);
			WifiUpdateConfig((ConfigNetwork*)pData);
		}
		break;
		case IC_PRESENTER_CFG_MISC:
		{
			ConfigMisc* pMisc = (ConfigMisc*)pData;
			MainStatusHwIdUpdate(pMisc->hardwareId);
			FeaturesSet(pMisc->features);
		}
		break;
		case IC_PRESENTER_WIFI_STAT:
		{
			memcpy(&presenter.lastWifiStat, pData, sizeof(PresenterWifiStat));

			WifiUpdateStatus((PresenterWifiStat*)pData);
		}
		break;
	}
}

static void kdCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	if(pHeader->command == IC_KD_FEEDBACK)
	{
		KdFeedback* pFb = (KdFeedback*)pData;

		if(chVTTimeElapsedSinceX(presenter.lastKdTime) > MS2ST(100))
		{
			MainStatusKdUpdate(pFb);
			presenter.lastKdTime = chVTGetSystemTime();
		}
	}
}

static void usbHcdCallback(USBHCDConnectionEvent* pConEvent)
{
	LoggingHCDUpdate(pConEvent);
}

static void usbMscCallback(USBMSCEvent* pMscEvent)
{
	LoggingMSCUpdate(pMscEvent);

	if(pMscEvent->connect)
	{
//		if(f_stat("autorun_bootloader", 0) == FR_OK)
//			ProgStoreStart(PROG_STORE_SOURCE_USB);
//		else
			LogFileOpen(0);
	}
	else
		LogFileClose();
}

static void logFileCallback(LogFileEvent* pLogEvent)
{
	LoggingLogUpdate(pLogEvent);

	if(pLogEvent->eventType == LOGFILE_EVENT_START)
	{
		ConsolePrint("Logfile started: %s\r\n", logFile.logFilename);
		NetworkPrint("Logfile started: %s\n", logFile.logFilename);
	}

	if(pLogEvent->eventType == LOGFILE_EVENT_CLOSE)
	{
		ConsolePrint("Log file closed\r\n");
		NetworkPrint("Log file closed\n");
	}
}

void PresenterInit()
{
	chHeapObjectInit(&guiHeap, guiHeapBuffer, GUI_HEAP_SIZE);

	TFTInit();
	gfxInit();
	gdispSetOrientation(GDISP_ROTATE_PORTRAIT);

	// Set the widget defaults
	gwinSetDefaultFont(gdispOpenFont("fixed_7x14"));
	gwinSetDefaultStyle(&BlackWidgetStyle, FALSE);
	gwinSetDefaultBgColor(Black);
	gwinSetDefaultColor(White);
	gdispClear(Black);

	handles.hTopBar = TopBarCreate();
	handles.hWindows[GUI_WINDOW_STATUS] = MainStatusCreate();
	handles.hWindows[GUI_WINDOW_WIFI_SETTINGS] = WifiCreate();
	handles.hWindows[GUI_WINDOW_BOT_ID] = SelectBotIDCreate();
	handles.hWindows[GUI_WINDOW_LOGGING] = LoggingCreate();
	handles.hWindows[GUI_WINDOW_FEATURES] = FeaturesCreate();
//	handles.hWindows[GUI_WINDOW_OPTICAL] = OpticalCreate();
	handles.hWindows[GUI_WINDOW_TEST_DRIVE] = TestDriveCreate();
	handles.hWindows[GUI_WINDOW_TEST_DRIBBLER] = TestDribblerCreate();

	handles.hMenu = MenuCreate(handles.hWindows, 9);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(client);

	client.rxCallback = &presenterCallback;
	client.section = INTERCOM_SECTION_PRESENTER;
	IntercomAddClient(client);

	usbHcd.pConCb = &usbHcdCallback;
	usbMsc.eventCb = &usbMscCallback;
	logFile.eventCb = &logFileCallback;
}

void PresenterTask(void* params)
{
	(void)params;

	chRegSetThreadName("Presenter");

	GListener gl;

	geventListenerInit(&gl);
	gwinAttachListener(&gl);

//	geventAttachSource(&gl, ginputGetMouse(0), GLISTEN_TOUCHDOWNMOVES);

	int16_t currentWindow = GUI_WINDOW_STATUS;

	TopBarSetTitle(((GWidgetObject*)handles.hWindows[currentWindow])->text);
	gwinSetVisible(handles.hWindows[currentWindow], TRUE);

	while(1)
	{
		GEvent* pEvent;

		pEvent = geventEventWait(&gl, TIME_INFINITE);

		if(TopBarIsMenuClicked(pEvent))
		{
			gwinSetVisible(handles.hTopBar, FALSE);
			gwinSetVisible(handles.hWindows[currentWindow], FALSE);

			gwinSetVisible(handles.hMenu, TRUE);

			continue;
		}

		int16_t menuSelection = -1;
		if(((GWidgetObject*)handles.hWindows[currentWindow])->fnParam)
		{
			// please close your eyes for the next line, looks messy ;)
			menuSelection = ((EventHandlerFunc)(((GWidgetObject*)handles.hWindows[currentWindow])->fnParam))(pEvent);
		}

		if(menuSelection < 0)
			menuSelection = MenuGetSelection(pEvent);

		if(menuSelection >= 0)
		{
			gwinSetVisible(handles.hWindows[currentWindow], FALSE);

			currentWindow = menuSelection;

			TopBarSetTitle(((GWidgetObject*)handles.hWindows[currentWindow])->text);

			gwinSetVisible(handles.hMenu, FALSE);
			gwinSetVisible(handles.hTopBar, TRUE);
			gwinSetVisible(handles.hWindows[currentWindow], TRUE);
		}
	}
}
