/*
 * drive.h
 *
 *  Created on: 02.02.2014
 *      Author: AndreR
 */

#ifndef DRIVE_H_
#define DRIVE_H_

#include <stdint.h>

#include "stm32f30x.h"
#include "ch.h"
#include "intercom_constants.h"
#include "util/flash_fs.h"
#include "arm_math.h"
#include "util/kf.h"
#include "util/luenberger.h"
#include "util/pid.h"
#include "util/abg_filter.h"

#define IN_CCM __attribute__((section(".ccmtext")))
#define IN_CCM_VAR __attribute__((section(".ccm")))

#define MOTOR_RESISTANCE	1.2f				// [Ohm]
#define MOTOR_KT			0.0255f				// [Nm/A]
#define MOTOR_KE			0.025532878570892f	// [V/rad]

#define DRIVE_ENC_SAMPLES 4
#define DRIVE_ADC_SAMPLES 16
#define DRIVE_MOVE_THRESHOLD 0.25f

#define DRIVE_MODE_DISABLE	0
#define DRIVE_MODE_BLOCK	1
#define DRIVE_MODE_SVPWM	2
#define DRIVE_MODE_TEST1	3
#define DRIVE_MODE_TEST2	4
#define DRIVE_MODE_TEST3	5

#define DRIVE_CTRL_MODE_NONE		0
#define DRIVE_CTRL_MODE_PID_SPEED	1
#define DRIVE_CTRL_MODE_PID_TORQUE	2

typedef struct _Drive
{
	const uint8_t id;

	// ### Setup ###
	// Motor and encoder timer
	TIM_TypeDef* const pTIMMotor;
	TIM_TypeDef* const pTIMEncoder;

	// ADC calibration values
	uint32_t adcCal[2];

	// ### Measured values ###
	// ADC DMA storage area
	union
	{
		uint16_t adc[2][DRIVE_ADC_SAMPLES];
		uint32_t adc32[2][DRIVE_ADC_SAMPLES/2];
	};
	
	uint16_t encDtA[DRIVE_ENC_SAMPLES];
	uint8_t encIndex;
	uint16_t lastDir;

	float rmsCurrent;
	uint32_t rmsCurrentTimestamp;

	uint16_t halPos;

	// ### control related ###
	struct _speed
	{
		PID pid;
		float setVelocity;
	} speed;

	struct _torque
	{
		PID pid;
		float setCurrent;
	} torque;

	float setVoltage;
	uint32_t setVoltageTimestamp;

	volatile uint8_t mode;

	volatile int32_t outLevel;	// [-3600 - 3600]

	volatile uint8_t ctrlMode;

	volatile uint8_t slopeLimited;

	// ### State Estimation ###
	float curVelocity;	// current velocity [rounds per second]
	uint32_t curVelocityTimestamp;

	float estCurrent;
	float estTorque;

	// ### required input ###
	float bat;	// [V]
	float cur;	// [A]
} Drive;

extern Drive drive[2];
extern Drive* const pRear;
extern Drive* const pFront;

void DriveInit();
void DriveTask(void* params);

// Low-level functions
void DriveSetMode(Drive* pDrive, uint8_t mode);
void DriveSetVoltage(Drive* pDrive, float voltage);

// Select controller
void DriveSetCtrlMode(Drive* pDrive, uint8_t mode);

// Speed controller
void DriveSpeedSetPIDParams(Drive* pDrive, float p, float i, float d);
void DriveSpeedSetVelocity(Drive* pDrive, float vel);

// Torque controller
void DriveTorqueSetPIDParams(Drive* pDrive, float p, float i, float d);
void DriveTorqueSetTorque(Drive* pDrive, float torque);

// SVPWM helper functions (mostly for tests)
void DriveSetPhaseVoltagesSVPWM(Drive* pDrive, float Va, float Vb, float Vc);
void DriveSetAlphaBeta(Drive* pDrive, float Valpha, float Vbeta);
void DriveSetVoltageAngle(Drive* pDrive, float Vref, float angle);

#endif /* DRIVE_H_ */
