/*
 * ctrl.h
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#ifndef CTRL_H_
#define CTRL_H_

#include <main/util.h>
#include "commands.h"
#include "arm_math.h"
#include "util/config.h"
#include "util/ema_filter.h"
#include "util/lag_element.h"
#include "log_msgs.h"
#include "intercom_constants.h"

#define CTRL_MODE_LOCAL_VEL		0
#define CTRL_MODE_GLOBAL_POS	2
#define CTRL_MODE_WHEEL_VEL		3
#define CTRL_MODE_OFF			5

#define MOT_CTRL_OFF			0
#define MOT_CTRL_VELOCITY		1
#define MOT_CTRL_VOLTAGE		2
#define MOT_CTRL_TORQUE			3

#define CTRL_BUZZ_UP20			0
#define CTRL_BUZZ_BEEP_FAST		1
#define CTRL_BUZZ_UP50			2
#define CTRL_BUZZ_DOWN50		3
#define CTRL_BUZZ_UPDOWN20		4

// helper for better readability
#define CTRL_MOTOR_TO_WHEEL_RATIO botParams.driveTrain.motor2WheelRatio
#define CTRL_WHEEL_TO_MOTOR_RATIO botParams.driveTrain.wheel2MotorRatio
#define CTRL_MOTOR_KM botParams.driveTrain.motor.Km
#define CTRL_MOTOR_KN botParams.driveTrain.motor.Kn
#define CTRL_MOTOR_KE botParams.driveTrain.motor.Ke
#define CTRL_MOTOR_KF botParams.driveTrain.motor.Kf
#define CTRL_MOTOR_R botParams.driveTrain.motor.R
#define CTRL_MOTOR_L botParams.driveTrain.motor.L

typedef LogCtrlInput CtrlInput;
typedef LogCtrlOutput CtrlOutput;
typedef LogCtrlState CtrlState;

typedef void(*CtrlInitFunction)();
typedef void(*CtrlStateFunction)(CtrlInput*, const CtrlOutput*);
typedef void(*CtrlFunction)(CtrlInput*, CtrlOutput*);
typedef void(*CtrlExitFunction)();
typedef void(*CtrlInputFunction)(uint8_t mode, float* pCmd);

typedef struct _CtrlInstance
{
	CtrlInitFunction pInit;
	CtrlStateFunction pStateEstimationFunc;
	CtrlFunction pControllerFunc;
	CtrlExitFunction pExit;
	CtrlInputFunction pInput;

	const float deltaT; // in [s], min. accuracy is 0.5ms
} CtrlInstance;

// Drive Train parameters
typedef struct PACKED _DriveTrainParams
{
	float motor2WheelRatio;
	float wheel2MotorRatio;
	struct PACKED
	{
		float Km;	// motor torque constant [Nm/A]
		float Kn;	// motor speed constant [rad/(s*V)]
		float Ke;	// motor back-EMF constant [V/rad]
		float Kf;	// motor friction constant [Nm*s]
		float R;	// motor resistance [Ohm]
		float L;	// motor inductance [H]
	} motor;
} DriveTrainParams;

// physical bot params
typedef struct PACKED _PhysicalParams
{
	float wheelRadius;
	float frontAngle;
	float backAngle;
	float botRadius;
	float mass; // [kg]
} PhysicalParams;

typedef struct _DriveLimits
{
	float velMaxXY;
	float accMaxXY;
	float jerkMaxXY;

	float velMaxW;
	float accMaxW;
	float jerkMaxW;
} DriveLimits;

typedef struct PACKED _BotParameters
{
	const DriveTrainParams driveTrain;
	PhysicalParams physical;
} BotParameters;

extern BotParameters botParams;

typedef struct PACKED _BotLogging
{
	uint8_t ctrlCommon;
	uint8_t ctrlSpecific;
} BotLogging;

typedef struct _Ctrl
{
	CtrlInput input;
	CtrlState state;
	CtrlOutput output;

	DriveLimits limits;

	LogKickerDribbler kd;
	LogCtrlDriveTrain logCfgDriveTrain;

	ConfigMisc misc;
	FlashFile* pMiscConfig;

	uint8_t feedbackStateUpdated[2];

	CtrlInstance* pInstance;
	CtrlInstance* pNewInstance;

	ConfigFile* pBotParamsConfig;
	ConfigFile* pBotLimitsConfig;
	BotLogging botLogging;
	ConfigFile* pBotLoggingConfig;
	float momentOfInertia;
	float theta[4];	// wheel angles in [rad]
	CtrlUtilPerfConstants perfConstants;

	arm_matrix_instance_f32 matXYW2Motor; // 4x3
	arm_matrix_instance_f32 matMotor2XYW; // 3x4

	arm_matrix_instance_f32 matForceXYW2Motor; // 4x3
	arm_matrix_instance_f32 matForceMotor2XYW; // 3x4

	arm_matrix_instance_f32	matMotorNullDir; // 4x1, linear combinations of this motor vector do not produce any movement = null space of matMotor2XYW

	arm_matrix_instance_f32 matForceXYZ2DownForce; // 4x3

	arm_matrix_instance_f32 matPos;
	arm_matrix_instance_f32 matVel;

	uint32_t logInterval;
	uint32_t lastLogTime;
	uint32_t lastPerfTime;
	uint32_t lastKdLogTime;
	uint32_t lastParamsLogTime;

	float defaultPosition[3];
	uint32_t lastDefaultPosStartTime;

	LagElementPT1 accZLag;
	LagElementPT1 gyrWLag;

	float accZ;
	float gyrW;

	EMAFilter avgCtrlRuntime;
	float cpuUsage; // percent, 0 - 100%

	uint8_t motorsOn;
	uint8_t buzzOnTimeViolation;

	systime_t ctrlIntervalUSec;
} Ctrl;

extern Ctrl ctrl;

void CtrlInit();
void CtrlTask(void* params);
void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass);
void CtrlSetController(CtrlInstance* pInstance);
void CtrlSetLogRate(uint32_t rate);
void CtrlSetMotorsOn(uint8_t enable);

void CtrlUpdateBotId(uint8_t botId);
void CtrlSetHardwareId(uint8_t hwId);
void CtrlSetFeatures(uint16_t features);

// this must be implemented by hardware specific ctrl implementation
void CtrlImplInit();
void CtrlImplUpdateInput(CtrlInput* pInput, LogKickerDribbler* pKD);
void CtrlImplApplyOutput(const CtrlOutput* pOutput);
void CtrlImplWriteLogMsg(void* pMsg);
void CtrlImplSendNetworkPacket(PacketHeader* pHeader, uint8_t* pData, uint16_t dataLength);
void CtrlImplKickerFire(uint8_t mode, uint8_t device, float speed);
void CtrlImplBuzzerPlay(uint16_t seqId);
void CtrlImplCorrectOutXYW(float* pLocalVel);
void CtrlImplCorrectInXYW(float* pLocalVel);

#endif /* CTRL_H_ */
