/*
 * ctrl_tigga.c
 *
 *  Created on: 26.02.2015
 *      Author: AndreR
 */

#include <main/ctrl.h>
#include <main/ctrl_tigga.h>
#include "struct_ids.h"
#include "util/arm_mat_util_f32.h"
#include "util/angle_math.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "util/console.h"

// DEFINITIONS
#define CTRL_TIGGA_DELTA_T_IN_US 6250
#define DATA_STORAGE_COVER_TIME_IN_US 200000

// FORWARD DECLARATIONS
static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);
static void sensorFusion(CtrlInput* pInput, const CtrlOutput* pOutput);
static void input(uint8_t mode, float* pCmd);

// GLOBAL DATA
CtrlTigga ctrlTigga = {
	.configFilter = {
		.state = {
			.posXY = 0.001f,
			.posW = 0.00001f,
			.velXY = 0.01f,
			.velW = 0.0001f,
			.accXY = 1.0f,
		},
		.sensor = {
			.visXY = 0.005f,
			.visW = 0.1f,
			.gyrW = 0.01f,
			.accXY = 100.0f,
		},
	},
	.configMisc = {
		.visionDelay = 30,
		.bsDelayMax = 20,
		.ASR = 0,
		.encoderAlpha = 0.9f,
		.dampingXY = 0.25f,
		.dampingW = 0.05f,
		.tShiftXY = 0.59f,
		.tShiftW = 0.59f,
		.useFeedforward = 1,
	},
	.configPid = {
		.posXY = {5.0f, 0, 0.0f},
		.posW = {10.0, 0, 0.0},
		.velXY = {0, 0, 0},
		.velW = {0, 0, 0},
	},
};

CtrlInstance ctrlTiggaInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &sensorFusion,
	.pControllerFunc = &controller,
	.pInput = &input,
	.deltaT = CTRL_TIGGA_DELTA_T_IN_US*1e-6f };

// LOCAL DATA - CONFIGS
static const ConfigFileDesc configFileDescFilter =
	{ SID_CFG_CTRL_TIGGA_FILTER, 23, "ctrl_tigga/filter", 9, (ElementDesc[]) {
		{ FLOAT, "state_pos_xy", "", "state/pos/xy" },
		{ FLOAT, "state_pos_w", "", "state/pos/w" },
		{ FLOAT, "state_vel_xy", "", "state/vel/xy" },
		{ FLOAT, "state_vel_w", "", "state/vel/w" },
		{ FLOAT, "state_acc_xy", "", "state/acc/xy" },
		{ FLOAT, "sensor_vision_xy", "", "sensor/vision/xy" },
		{ FLOAT, "sensor_vision_w", "", "sensor/vision/w" },
		{ FLOAT, "sensor_gyro_w", "", "sensor/gyro/w" },
		{ FLOAT, "sensor_acc_xy", "", "sensor/acc/xy" },
	} };

static const ConfigFileDesc configFileDescMisc =
	{ SID_CFG_CTRL_TIGGA_MISC, 7, "ctrl_tigga/misc", 9, (ElementDesc[]) {
		{ UINT16, "vision_delay", "ms", "visionDelay" },
		{ UINT16, "bs_delay", "ms", "bsDelay" },
		{ UINT16, "asr", "", "ASR" },
		{ FLOAT, "enc_alpha", "", "Encoder adj. rate" },
		{ FLOAT, "damping_xy", "", "Traj. damping (XY)" },
		{ FLOAT, "damping_w", "", "Traj. damping (W)" },
		{ FLOAT, "tshift_xy", "", "Traj. shifting (XY)" },
		{ FLOAT, "tshift_w", "", "Traj. shifting (W)" },
		{ UINT8, "feedforward", "", "Use Feedforward" },
	} };

static const ConfigFileDesc configFileDescPid =
	{ SID_CFG_CTRL_TIGGA_PID, 1, "ctrl_tigga/pid", 12, (ElementDesc[]) {
		{ FLOAT, "pos_xy_p", "", "pos/xy/p" },
		{ FLOAT, "pos_xy_i", "", "pos/xy/i" },
		{ FLOAT, "pos_xy_d", "", "pos/xy/d" },
		{ FLOAT, "pos_w_p", "", "pos/w/p" },
		{ FLOAT, "pos_w_i", "", "pos/w/i" },
		{ FLOAT, "pos_w_d", "", "pos/w/d" },
		{ FLOAT, "vel_xy_p", "", "vel/xy/p" },
		{ FLOAT, "vel_xy_i", "", "vel/xy/i" },
		{ FLOAT, "vel_xy_d", "", "vel/xy/d" },
		{ FLOAT, "vel_w_p", "", "vel/w/p" },
		{ FLOAT, "vel_w_i", "", "vel/w/i" },
		{ FLOAT, "vel_w_d", "", "vel/w/d" },
	} };

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case SID_CFG_CTRL_TIGGA_FILTER:
		{
			ctrlTigga.stateNoise[0] = ctrlTigga.configFilter.state.posXY;
			ctrlTigga.stateNoise[1] = ctrlTigga.configFilter.state.posXY;
			ctrlTigga.stateNoise[2] = ctrlTigga.configFilter.state.posW;
			ctrlTigga.stateNoise[3] = ctrlTigga.configFilter.state.velXY;
			ctrlTigga.stateNoise[4] = ctrlTigga.configFilter.state.velXY;
			ctrlTigga.stateNoise[5] = ctrlTigga.configFilter.state.velW;
			ctrlTigga.stateNoise[6] = ctrlTigga.configFilter.state.accXY;
			ctrlTigga.stateNoise[7] = ctrlTigga.configFilter.state.accXY;

			ctrlTigga.obsNoise[0] = ctrlTigga.configFilter.sensor.visXY;
			ctrlTigga.obsNoise[1] = ctrlTigga.configFilter.sensor.visXY;
			ctrlTigga.obsNoise[2] = ctrlTigga.configFilter.sensor.visW;
			ctrlTigga.obsNoise[3] = ctrlTigga.configFilter.sensor.gyrW;
			ctrlTigga.obsNoise[4] = ctrlTigga.configFilter.sensor.accXY;
			ctrlTigga.obsNoise[5] = ctrlTigga.configFilter.sensor.accXY;

			SRUKFSetProcessNoise(&ctrlTigga.ukf, ctrlTigga.stateNoise);
			SRUKFSetMeasurementNoise(&ctrlTigga.ukf, ctrlTigga.obsNoise);
		}
		break;
		case SID_CFG_CTRL_TIGGA_PID:
		{
			PIDSetParams(&ctrlTigga.posPIDXY, ctrlTigga.configPid.posXY.kp, ctrlTigga.configPid.posXY.ki, ctrlTigga.configPid.posXY.kd);
			PIDSetParams(&ctrlTigga.posPIDW, ctrlTigga.configPid.posW.kp, ctrlTigga.configPid.posW.ki, ctrlTigga.configPid.posW.kd);

			PIDSetParams(&ctrlTigga.velPIDXY, ctrlTigga.configPid.velXY.kp, ctrlTigga.configPid.velXY.ki, ctrlTigga.configPid.velXY.kd);
			PIDSetParams(&ctrlTigga.velPIDW, ctrlTigga.configPid.velW.kp, ctrlTigga.configPid.velW.ki, ctrlTigga.configPid.velW.kd);
		}
		break;
		case SID_CFG_CTRL_TIGGA_MISC:
		{
			ctrlTigga.trajGen.dampingXY = ctrlTigga.configMisc.dampingXY;
			ctrlTigga.trajGen.dampingW = ctrlTigga.configMisc.dampingW;
			ctrlTigga.trajGen.tShiftXY = ctrlTigga.configMisc.tShiftXY;
			ctrlTigga.trajGen.tShiftW = ctrlTigga.configMisc.tShiftW;

			ctrlTigga.encFilter[0].alpha = ctrlTigga.configMisc.encoderAlpha;
			ctrlTigga.encFilter[1].alpha = ctrlTigga.configMisc.encoderAlpha;
		}
		break;
	}
}

static void stateFunc(arm_matrix_instance_f32* _pX, const arm_matrix_instance_f32* _pU)
{
	// state vector: [px py pw vx vy vw ax ay]
	// control vector: [ax ay aw]
	// velocity/acceleration in local frame

	// PT1 constant gain factors for X and Y
	const float K[2] = {0.749f, 0.89f};
	// PT1 time constants for X and Y
	const float T[2] = {0.15f, 0.08f};

	const float dT = ctrlTiggaInstance.deltaT;

	const float T2[2] = {1.0f/((T[0]/dT)+1.0f), 1.0f/((T[1]/dT)+1.0f)};

	float* pX = _pX->pData;
	float* pU = _pU->pData;

	float pos[3] = {pX[0], pX[1], pX[2]};
	float velLocal[3] = {pX[3], pX[4], pX[5]};
	float accLocal[2] = {pX[6], pX[7]};

	pX[6] = T2[0]*(K[0]*pU[0]-accLocal[0]) + accLocal[0];
	pX[7] = T2[1]*(K[1]*pU[1]-accLocal[1]) + accLocal[1];

	pX[3] = velLocal[0] + accLocal[0]*dT;
	pX[4] = velLocal[1] + accLocal[1]*dT;
	pX[5] = velLocal[2] + pU[2]*dT;

	pX[2] = pos[2] + velLocal[2]*dT + 0.5f*pU[2]*dT*dT;

	float beta = -PI/2.0f + pos[2];
	float sinA = sinf(beta);
	float cosA = cosf(beta);

	float velGlobal[2];
	velGlobal[0] = cosA*velLocal[0] - sinA*velLocal[1];
	velGlobal[1] = cosA*velLocal[1] + sinA*velLocal[0];

	pX[0] = pos[0] + velGlobal[0]*dT;
	pX[1] = pos[1] + velGlobal[1]*dT;
}

static void measFunc(const arm_matrix_instance_f32* _pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* _pY)
{
	(void)pU;

	// observation vector: [px py pw vw ax ay]

	float* pY = _pY->pData;
	float* pX = _pX->pData;

	pY[0] = pX[0];
	pY[1] = pX[1];
	pY[2] = pX[2];
	pY[3] = pX[5];

//	pY[4] = pX[6];
//	pY[5] = pX[7];
	pY[4] = pX[6] - pX[4]*pX[5];
	pY[5] = pX[7] + pX[3]*pX[5];
}

static void init()
{
	const float dt = ctrlTiggaInstance.deltaT;

	ctrlTigga.log.header.type = SID_CTRL_TIGGA_DEBUG;
	ctrlTigga.log.header.length = sizeof(LogCtrlTiggaDebug);

	PIDParams pp;
	pp.kp = 1.0f;
	pp.kd = 0;
	pp.ki = 0;
	pp.outputMin = -100;
	pp.outputMax = 100;
	PIDInit(&ctrlTigga.velPIDW, &pp, CTRL_TIGGA_DELTA_T_IN_US*1e-6f, 1);

	pp.outputMin = -5.0f;
	pp.outputMax = 5.0f;
	PIDInit(&ctrlTigga.velPIDXY, &pp, CTRL_TIGGA_DELTA_T_IN_US*1e-6f, 1);

	pp.outputMin = -5.0f;
	pp.outputMax = 5.0f;
	PIDInit(&ctrlTigga.posPIDXY, &pp, dt, 1);

	pp.outputMin = -50.0f;
	pp.outputMax = 50.0f;
	PIDInit(&ctrlTigga.posPIDW, &pp, dt, 1);

	// Square-Root Unscented Kalman Filter
	// state vector: [px py pw vx vy vw ax ay]
	// sensor vector: [px_vis py_vis pw_vis vw_gyr ax_acc ay_acc]
	// control vector: [ax ay aw]
	SRUKFInit(&ctrlTigga.ukf, 8, 6, 3, ctrlTigga.ukfData, &stateFunc, &measFunc, ctrlTigga.stateNoise, ctrlTigga.obsNoise);
	ctrlTigga.ukf.pAngularMeas[2] = 1;
	ctrlTigga.ukf.pAngularState[2] = 1;

	TrajGeneratorInit(&ctrlTigga.trajGen);

	EMAFilterInit(&ctrlTigga.encFilter[0], 0.9f);
	EMAFilterInit(&ctrlTigga.encFilter[1], 0.9f);
	ctrlTigga.encFilter[0].value = 1.0f;
	ctrlTigga.encFilter[1].value = 1.0f;

	ConfigNotifyUpdate(ctrlTigga.pConfigFileFilter);
	ConfigNotifyUpdate(ctrlTigga.pConfigFilePid);
	ConfigNotifyUpdate(ctrlTigga.pConfigFileMisc);
}

static inline void updatePD2D(PID* pPID, const float* pPosSet, const float* pPosMeas, const float* pVelSet, const float* pVelMeas, float* pOut)
{
	// XY position control
	float errorPosX = pPosMeas[0] - pPosSet[0];
	float errorPosY = pPosMeas[1] - pPosSet[1];
	float errorPosXY = sqrtf(errorPosX*errorPosX+errorPosY*errorPosY);

	// XY velocity control
	float errorVelX = pVelMeas[0] - pVelSet[0];
	float errorVelY = pVelMeas[1] - pVelSet[1];
	float errorVelXY = sqrtf(errorVelX*errorVelX+errorVelY*errorVelY);

	if(errorPosXY > 0.0001f)
	{
		pOut[0] = errorPosX/errorPosXY * -pPID->params.kp;
		pOut[1] = errorPosY/errorPosXY * -pPID->params.kp;
	}
	else
	{
		pOut[0] = 0.0f;
		pOut[1] = 0.0f;
	}

	if(errorVelXY > 0.0001f)
	{
		pOut[0] += errorVelX/errorVelXY * -pPID->params.kd * pPID->sampleTime;
		pOut[1] += errorVelY/errorVelXY * -pPID->params.kd * pPID->sampleTime;
	}
}

static inline void updatePID2D(PID* pPID, const float* pSet, float* pMeas, float* pOut, uint8_t firstRun)
{
	// XY position control
	float errorPosX = pMeas[0] - pSet[0];
	float errorPosY = pMeas[1] - pSet[1];
	float errorXY = sqrtf(errorPosX*errorPosX+errorPosY*errorPosY);

	// process PIDs
	pPID->setpoint = 0;
	if(firstRun)
		pPID->lastMeasurement = errorXY; // prevent derivative kick
	PIDUpdate(pPID, errorXY);

	if(errorXY > 0.0001f)
	{
		pOut[0] = errorPosX/errorXY * pPID->output;
		pOut[1] = errorPosY/errorXY * pPID->output;
	}
	else
	{
		pOut[0] = 0.0f;
		pOut[1] = 0.0f;
	}

	if(errorXY > 0.5f)
	{
		TrajGeneratorReset(&ctrlTigga.trajGen);
	}
}

static inline void modelFuncLocalInv(const float* pVel, const float* pAcc, const float* pJerk, float tStep, float* pOutAcc)
{
	// PT1 constant gain factors for X and Y {acc, brk}
	const float Kx[2] = {0.6461f, 0.9648f};
	const float Ky[2] = {0.8377f, 1.0140f};

	// PT1 time constants, first-order poly dependent on absolute speed
	// T = T1*|v|+T0
	const float T1x[2] = {0.0672f, 0.1021f};
	const float T0x[2] = {0.0193f, 0.0121f};
	const float T1y[2] = {0.0236f, 0.0467f};
	const float T0y[2] = {0.0195f, 0.0191f};

	float vNext[2] = {pVel[0]+pAcc[0]*tStep, pVel[1]+pAcc[1]*tStep};

	float absVNext = sqrtf(vNext[0]*vNext[0]+vNext[1]*vNext[1]);
	float absVel = sqrtf(pVel[0]*pVel[0]+pVel[1]*pVel[1]);

	float K[2];
	float T[2];

	if(absVNext > absVel)
	{
		// acceleration case
		K[0] = Kx[0];
		K[1] = Ky[0];

		T[0] = absVel*T1x[0] + T0x[0];
		T[1] = absVel*T1y[0] + T0y[0];
	}
	else
	{
		// deceleration case
		K[0] = Kx[1];
		K[1] = Ky[1];

		T[0] = absVel*T1x[1] + T0x[1];
		T[1] = absVel*T1y[1] + T0y[1];
	}

	pOutAcc[0] = 1.0f/K[0]*pAcc[0] + T[0]/K[0]*pJerk[0];
	pOutAcc[1] = 1.0f/K[1]*pAcc[1] + T[1]/K[1]*pJerk[1];
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	(void)pInput;

	uint32_t startTime = SysTimeUSecRaw();

	const float dT = ctrlTiggaInstance.deltaT;

	float curPos[3] = {0, 0, 0};
	float curVel[3] = {0, 0, 0};

	// copy state to local variable to prevent unaligned access faults
	memcpy(curPos, ctrl.state.pos, sizeof(float)*3);
	memcpy(curVel, ctrl.state.vel, sizeof(float)*3);

	// Configure trajectory generator
	ctrlTigga.trajGen.vMaxXY = ctrl.limits.velMaxXY;
	ctrlTigga.trajGen.vMaxW = ctrl.limits.velMaxW;
	ctrlTigga.trajGen.aMaxXY = ctrl.limits.accMaxXY;
	ctrlTigga.trajGen.aMaxW = ctrl.limits.accMaxW;
	ctrlTigga.trajGen.jMaxXY = ctrl.limits.jerkMaxXY;
	ctrlTigga.trajGen.jMaxW = ctrl.limits.jerkMaxW;

	TrajGeneratorSetState(&ctrlTigga.trajGen, curPos, curVel);

	uint8_t ctrlFirstRun = 0;

	if(ctrl.motorsOn == 0)
		ctrlTigga.mode = CTRL_MODE_OFF;

	if(ctrlTigga.mode != ctrlTigga.lastMode)
	{
		ctrlTigga.lastMode = ctrlTigga.mode;

		ctrlFirstRun = 1;

		switch(ctrlTigga.mode)
		{
			case CTRL_MODE_LOCAL_VEL:
			{
			}
			break;
			case CTRL_MODE_GLOBAL_POS:
			{
			}
			break;
		}
	}

	switch(ctrlTigga.mode)
	{
		case CTRL_MODE_GLOBAL_POS:
		{
			// Trajectory Update
			TrajGeneratorCreateGlobalPos(&ctrlTigga.trajGen, ctrlTigga.setpointPos);

			float posNow[3];
			memcpy(posNow, ctrlTigga.trajGen.trajGlobalPos, sizeof(float)*3);

			TrajGeneratorStep(&ctrlTigga.trajGen, dT);

			// do some transformations for logging
			const float* trgPos = ctrlTigga.trajGen.trajGlobalPos;
			const float* trgVel = ctrlTigga.trajGen.trajGlobalVel;
			const float* trgAcc = ctrlTigga.trajGen.trajGlobalAcc;
			const float* trgJerk = ctrlTigga.trajGen.trajGlobalJerk;

			float trgVelLocal[3] = {0, 0, trgVel[2]};
			CtrlUtilTurnGlobal2Local(trgPos[2], trgVel[0], trgVel[1], trgVelLocal, trgVelLocal+1);

			float trgAccLocal[3] = {0, 0, trgAcc[2]};
			CtrlUtilTurnGlobal2Local(trgPos[2], trgAcc[0], trgAcc[1], trgAccLocal, trgAccLocal+1);

			float trgJerkLocal[3] = {0, 0, trgJerk[2]};
			CtrlUtilTurnGlobal2Local(trgPos[2], trgJerk[0], trgJerk[1], trgJerkLocal, trgJerkLocal+1);

			memcpy(ctrlTigga.log.trajPos, trgPos, sizeof(float)*3);
			memcpy(ctrlTigga.log.trajVelLocal, trgVelLocal, sizeof(float)*3);
			memcpy(ctrlTigga.log.trajAccLocal, trgAccLocal, sizeof(float)*3);
			memcpy(ctrlTigga.log.trajJerkLocal, trgJerkLocal, sizeof(float)*3);

			// XY position control
			float pidOutputXY[2];
//			updatePD2D(&ctrlTigga.posPIDXY, posNow, curPos, velNow, curVel, pidOutputXY);
			updatePID2D(&ctrlTigga.posPIDXY, posNow, curPos, pidOutputXY, ctrlFirstRun);

			// orientation processing
			ctrlTigga.posPIDW.setpoint = posNow[2];
			if(ctrlFirstRun)
				ctrlTigga.posPIDW.lastMeasurement = curPos[2];
			PIDUpdateAngular(&ctrlTigga.posPIDW, curPos[2]);

			// convert to local pos PID output
			float localPosPidOutput[3] = {0, 0, ctrlTigga.posPIDW.output};
			CtrlUtilTurnGlobal2Local(curPos[2], pidOutputXY[0], pidOutputXY[1], localPosPidOutput, localPosPidOutput+1);

			// slope limiter
			static float lastPidVel[3] = {0, 0, 0};
			if(ctrlFirstRun)
				memcpy(lastPidVel, localPosPidOutput, sizeof(float)*3);

			float acc = (localPosPidOutput[2]-lastPidVel[2])/dT;
			if(acc > ctrl.limits.accMaxW*2)
				localPosPidOutput[2] = lastPidVel[2]+ctrl.limits.accMaxW*dT*2;
			if(acc < -ctrl.limits.accMaxW*2)
				localPosPidOutput[2] = lastPidVel[2]-ctrl.limits.accMaxW*dT*2;

			for(uint8_t i = 0; i < 2; i++)
			{
				acc = (localPosPidOutput[i]-lastPidVel[i])/dT;
				if(acc > ctrl.limits.accMaxXY*2)
					localPosPidOutput[i] = lastPidVel[i]+ctrl.limits.accMaxXY*dT*2;
				if(acc < -ctrl.limits.accMaxXY*2)
					localPosPidOutput[i] = lastPidVel[i]-ctrl.limits.accMaxXY*dT*2;
			}

			memcpy(lastPidVel, localPosPidOutput, sizeof(float)*3);

			// feedforward control by model-inversion
			float accFF[2] = {0, 0};
			if(ctrlTigga.configMisc.useFeedforward)
				modelFuncLocalInv(trgVelLocal, trgAccLocal, trgJerkLocal, dT, accFF);

			float outVel[3];
			outVel[0] = localPosPidOutput[0] + trgVelLocal[0] + accFF[0]*dT;
			outVel[1] = localPosPidOutput[1] + trgVelLocal[1] + accFF[1]*dT;
			outVel[2] = localPosPidOutput[2] + trgVelLocal[2];

			CtrlUtilLocalVelToMotorVel(outVel, pOutput->motorVel, ctrlTigga.configMisc.ASR);
			pOutput->ctrlMode = MOT_CTRL_VELOCITY;
		}
		break;
		case CTRL_MODE_LOCAL_VEL:
		{
			// Trajectory Update
			TrajGeneratorCreateLocalVel(&ctrlTigga.trajGen, ctrlTigga.setpointVel);

			float velNow[3];
			memcpy(velNow, ctrlTigga.trajGen.trajLocalVel, sizeof(float)*3);

			TrajGeneratorStep(&ctrlTigga.trajGen, dT);

			// do some logging
			const float* trgVel = ctrlTigga.trajGen.trajLocalVel;
			const float* trgAcc = ctrlTigga.trajGen.trajLocalAcc;
			const float* trgJerk = ctrlTigga.trajGen.trajLocalJerk;

			memcpy(ctrlTigga.log.trajVelLocal, trgVel, sizeof(float)*3);
			memcpy(ctrlTigga.log.trajAccLocal, trgAcc, sizeof(float)*3);
			memcpy(ctrlTigga.log.trajJerkLocal, trgJerk, sizeof(float)*3);

			float localVel[3] = {0, 0, curVel[2]};
			CtrlUtilTurnGlobal2Local(curPos[2], curVel[0], curVel[1], localVel, localVel+1);

			// XY velocity control
			float pidOutputXY[2];
			updatePID2D(&ctrlTigga.velPIDXY, velNow, localVel, pidOutputXY, ctrlFirstRun);

			// W velocity control
			ctrlTigga.velPIDW.setpoint = velNow[2];
			if(ctrlFirstRun)
				ctrlTigga.velPIDW.lastMeasurement = localVel[2];
			PIDUpdate(&ctrlTigga.velPIDW, localVel[2]);

			// feedforward control by model-inversion
			float accFF[2] = {0, 0};
			if(ctrlTigga.configMisc.useFeedforward)
				modelFuncLocalInv(trgVel, trgAcc, trgJerk, dT, accFF);

			float outVel[3];
			outVel[0] = pidOutputXY[0] + velNow[0] + accFF[0]*dT;
			outVel[1] = pidOutputXY[1] + velNow[1] + accFF[1]*dT;
			outVel[2] = ctrlTigga.velPIDW.output + velNow[2] + trgAcc[2]*dT;

			CtrlUtilLocalVelToMotorVel(outVel, pOutput->motorVel, ctrlTigga.configMisc.ASR);
			pOutput->ctrlMode = MOT_CTRL_VELOCITY;
		}
		break;
		case CTRL_MODE_WHEEL_VEL:
		{
			TrajGeneratorReset(&ctrlTigga.trajGen);

			for(uint8_t i = 0; i < 4; i++)
				pOutput->motorVel[i] = ctrlTigga.setpointWheel[i]*CTRL_WHEEL_TO_MOTOR_RATIO;

			pOutput->ctrlMode = MOT_CTRL_VELOCITY;
		}
		break;
		default:
		{
			TrajGeneratorReset(&ctrlTigga.trajGen);

			memset(pOutput->motorVel, 0, sizeof(float)*4);

			pOutput->ctrlMode = MOT_CTRL_OFF;
		}
		break;
	}

	for(uint8_t i = 0; i < 3; i++)
	{
		ctrlTigga.log.setPos[i] = ctrlTigga.setpointPos[i];
		ctrlTigga.log.setVel[i] = ctrlTigga.setpointVel[i];
	}

	ctrlTigga.log.ctrlMode = ctrlTigga.mode;
	ctrlTigga.log.ctrlTime = SysTimeUSecRaw()-startTime;
	ctrlTigga.log.header.timestamp = SysTimeUSec();

	// send controller specific log data
	ctrlTigga.log.header.timestamp = SysTimeUSec();

	if(ctrl.botLogging.ctrlSpecific)
		CtrlImplWriteLogMsg(&ctrlTigga.log);
}

static void sensorFusion(CtrlInput* pInput, const CtrlOutput* pOutput)
{
	const float dT = ctrlTiggaInstance.deltaT;

	uint32_t startTime = SysTimeUSecRaw();

	uint8_t eventVisionUpdated = 0;
	uint8_t eventVisionRestored = 0;

	// process output, get commanded acceleration
	float localOutAcc[3] = {0, 0, 0};
	if(pOutput->ctrlMode == MOT_CTRL_VELOCITY)
	{
		float localOutVel[3];
		CtrlUtilMotorVelToLocalVel(pOutput->motorVel, localOutVel);

		localOutAcc[0] = (localOutVel[0]-ctrlTigga.lastLocalVelOut[0])/dT;
		localOutAcc[1] = (localOutVel[1]-ctrlTigga.lastLocalVelOut[1])/dT;
		localOutAcc[2] = (localOutVel[2]-ctrlTigga.lastLocalVelOut[2])/dT;

		memcpy(ctrlTigga.lastLocalVelOut, localOutVel, sizeof(float)*3);
	}
	else
	{
		memset(ctrlTigga.lastLocalVelOut, 0, sizeof(float)*3);
	}

	// set SR-UKF input
	MAT_ELEMENT(ctrlTigga.ukf.u, 0, 0) = localOutAcc[0];
	MAT_ELEMENT(ctrlTigga.ukf.u, 1, 0) = localOutAcc[1];
	MAT_ELEMENT(ctrlTigga.ukf.u, 2, 0) = localOutAcc[2];

	// predict one time step with internal model
	SRUKFPredict(&ctrlTigga.ukf);

	// default condition: no vision update
	ctrlTigga.ukf.pUpdate[0] = 0;
	ctrlTigga.ukf.pUpdate[1] = 0;
	ctrlTigga.ukf.pUpdate[2] = 0;

	// VISION
	if(pInput->vision.updated)
	{
		uint32_t timeNew = pInput->vision.time - pInput->vision.delay;
		uint32_t timeDiff = timeNew - ctrlTigga.lastVision.time;

		// at least two measurements with 10-100ms delay are needed for extrapolation
		if(timeDiff > 10000 && timeDiff < 100000)
		{
			float timeRate = 1.0f/(timeDiff*1e-6f);
			float extrapTime = (pInput->vision.delay+ctrlTigga.configMisc.visionDelay)*1e-6f;

			float extrapPos[3];
			for(uint8_t i = 0; i < 3; i++)
			{
				float vel = (pInput->vision.pos[i]-ctrlTigga.lastVision.pos[i])*timeRate;
				if(fabsf(vel) < 10.0f)	// sanity check
					extrapPos[i] = pInput->vision.pos[i] + vel*extrapTime;
				else
					extrapPos[i] = pInput->vision.pos[i];
			}

			ctrlTigga.ukf.pUpdate[0] = 1;
			ctrlTigga.ukf.pUpdate[1] = 1;
			ctrlTigga.ukf.pUpdate[2] = 1;

			MAT_ELEMENT(ctrlTigga.ukf.y, 0, 0) = extrapPos[0];
			MAT_ELEMENT(ctrlTigga.ukf.y, 1, 0) = extrapPos[1];
			MAT_ELEMENT(ctrlTigga.ukf.y, 2, 0) = extrapPos[2];

			eventVisionUpdated = 1;
		}

		memcpy(ctrlTigga.lastVision.pos, pInput->vision.pos, sizeof(float)*3);
		ctrlTigga.lastVision.time = timeNew;
		ctrlTigga.lastVision.camId = pInput->vision.camId;

		if(ctrlTigga.visionLost)
			eventVisionRestored = 1;

		ctrlTigga.lastVisionInputTime = chVTGetSystemTimeX();
		ctrlTigga.visionLost = 0;
	}
	else
	{
		if(chVTTimeElapsedSinceX(ctrlTigga.lastVisionInputTime) > MS2ST(100))
		{
			ctrlTigga.visionLost = 1;
			ctrlTigga.lastVisionInputTime = chVTGetSystemTimeX();
		}
	}

	// ENCODER
	float localVelEnc[3];
	CtrlUtilMotorVelToLocalVel(pInput->enc.vel, localVelEnc);

	if(pInput->vision.updated)
	{
		memcpy(ctrlTigga.encPosition, pInput->vision.pos, sizeof(float)*3);
	}
	else
	{
		// dead reckoning with encoders
		float localVenEncFilt[3] = {0, 0, localVelEnc[2]};
		localVenEncFilt[0] = localVelEnc[0]*ctrlTigga.encFilter[0].value;
		localVenEncFilt[1] = localVelEnc[1]*ctrlTigga.encFilter[1].value;

		float globalVelEnc[3] = {0, 0, localVenEncFilt[2]};
		CtrlUtilTurnLocal2Global(ctrlTigga.encPosition[2], localVenEncFilt[0], localVenEncFilt[1], globalVelEnc, globalVelEnc+1);

		ctrlTigga.encPosition[0] += globalVelEnc[0]*dT;
		ctrlTigga.encPosition[1] += globalVelEnc[1]*dT;
		ctrlTigga.encPosition[2] += globalVelEnc[2]*dT;

		AngleNormalizePtr(ctrlTigga.encPosition+2);
	}

	if(ctrlTigga.visionLost)
	{
		ctrlTigga.ukf.pUpdate[0] = 1;
		ctrlTigga.ukf.pUpdate[1] = 1;
		ctrlTigga.ukf.pUpdate[2] = 1;

		MAT_ELEMENT(ctrlTigga.ukf.y, 0, 0) = ctrlTigga.encPosition[0];
		MAT_ELEMENT(ctrlTigga.ukf.y, 1, 0) = ctrlTigga.encPosition[1];
		MAT_ELEMENT(ctrlTigga.ukf.y, 2, 0) = ctrlTigga.encPosition[2];
	}

	// GYRO
	MAT_ELEMENT(ctrlTigga.ukf.y, 3, 0) = pInput->gyr.rotVel[2];

	// ACCELEROMETER
	MAT_ELEMENT(ctrlTigga.ukf.y, 4, 0) = pInput->acc.linAcc[0];
	MAT_ELEMENT(ctrlTigga.ukf.y, 5, 0) = pInput->acc.linAcc[1];

	// all inputs set, perform SR-UKF update step
	SRUKFUpdate(&ctrlTigga.ukf);

	if(eventVisionRestored)
	{
		// vision restored, reset generator to make sure new filter position is used
		TrajGeneratorReset(&ctrlTigga.trajGen);
	}

	// update encoder mapping if a vision update was performed
	if(eventVisionUpdated)
	{
		float localFilterVel[2] = {ctrlTigga.ukf.x.pData[3], ctrlTigga.ukf.x.pData[4]};

		for(uint8_t i = 0; i < 2; i++)
		{
			if(fabsf(localFilterVel[i]) > 0.1f && fabsf(localVelEnc[i]) > 0.1f)
			{
				float enc2Real = fabsf(localFilterVel[i]/localVelEnc[i]);
				if(enc2Real <= 1.0f && enc2Real > 0.2f)
				{
					EMAFilterUpdate(&ctrlTigga.encFilter[i], enc2Real);
				}
			}

			ctrlTigga.log.enc2Real[i] = ctrlTigga.encFilter[i].value;
		}
	}

	memcpy(ctrl.state.pos, ctrlTigga.ukf.x.pData, sizeof(float)*3);

	float globalVel[3] = {0, 0, ctrlTigga.ukf.x.pData[5]};
	CtrlUtilTurnLocal2Global(ctrlTigga.ukf.x.pData[2], ctrlTigga.ukf.x.pData[3], ctrlTigga.ukf.x.pData[4], globalVel, globalVel+1);
	memcpy(ctrl.state.vel, globalVel, sizeof(float)*3);

	float globalAcc[2];
	CtrlUtilTurnLocal2Global(ctrlTigga.ukf.x.pData[2], ctrlTigga.ukf.x.pData[6], ctrlTigga.ukf.x.pData[7], globalAcc, globalAcc+1);
	ctrl.state.acc[0] = globalAcc[0];
	ctrl.state.acc[1] = globalAcc[1];
	ctrl.state.acc[2] = localOutAcc[2];

	ctrl.state.posUpdated = 1;
	ctrl.state.velUpdated = 1;
	ctrl.state.accUpdated = 1;

	for(uint8_t i = 0; i < 3; i++)
	{
		ctrlTigga.log.kfSigmaPos[i] = MAT_ELEMENT(ctrlTigga.ukf.Sx, i, i);
		ctrlTigga.log.kfSigmaVel[i] = MAT_ELEMENT(ctrlTigga.ukf.Sx, i+3, i+3);
	}

	ctrlTigga.log.kfSigmaAcc[0] = MAT_ELEMENT(ctrlTigga.ukf.Sx, 6, 6);
	ctrlTigga.log.kfSigmaAcc[1] = MAT_ELEMENT(ctrlTigga.ukf.Sx, 7, 7);

	ctrlTigga.log.fusionTime = SysTimeUSecRaw()-startTime;
}

static void input(uint8_t mode, float* pCmd)
{
	switch(mode)
	{
		case CTRL_MODE_LOCAL_VEL:
		{
			CtrlTiggaSetVelTarget(pCmd[0], pCmd[1], pCmd[2]);
		}
		break;
		case CTRL_MODE_GLOBAL_POS:
		{
			CtrlTiggaSetPosTarget(pCmd[0], pCmd[1], pCmd[2], pCmd[3]);
		}
		break;
		case CTRL_MODE_WHEEL_VEL:
		{
			CtrlTiggaSetWheelVelocity(pCmd[0], pCmd[1], pCmd[2], pCmd[3]);
		}
		break;
	}
}

void CtrlTiggaInitCfg()
{
	ctrlTigga.pConfigFileFilter = ConfigOpenOrCreate(&configFileDescFilter, &ctrlTigga.configFilter, sizeof(CtrlTiggaConfigFilter), &configUpdateCallback);
	ctrlTigga.pConfigFilePid = ConfigOpenOrCreate(&configFileDescPid, &ctrlTigga.configPid, sizeof(CtrlTiggaConfigPID), &configUpdateCallback);
	ctrlTigga.pConfigFileMisc = ConfigOpenOrCreate(&configFileDescMisc, &ctrlTigga.configMisc, sizeof(CtrlTiggaConfigMisc), &configUpdateCallback);
}

void CtrlTiggaSetVelTarget(float x, float y, float w)
{
	ctrlTigga.setpointVel[0] = x;
	ctrlTigga.setpointVel[1] = y;
	ctrlTigga.setpointVel[2] = w;

	ctrlTigga.mode = CTRL_MODE_LOCAL_VEL;
}

void CtrlTiggaSetPosTarget(float x, float y, float w, float t)
{
	(void)t;

	ctrlTigga.setpointPos[0] = x;
	ctrlTigga.setpointPos[1] = y;
	ctrlTigga.setpointPos[2] = w;

	ctrlTigga.setpointVel[0] = 0;
	ctrlTigga.setpointVel[1] = 0;
	ctrlTigga.setpointVel[2] = 0;

	ctrlTigga.mode = CTRL_MODE_GLOBAL_POS;
}

void CtrlTiggaSetWheelVelocity(float fr, float fl, float rl, float rr)
{
	ctrlTigga.setpointWheel[0] = fr;
	ctrlTigga.setpointWheel[1] = fl;
	ctrlTigga.setpointWheel[2] = rl;
	ctrlTigga.setpointWheel[3] = rr;

	ctrlTigga.mode = CTRL_MODE_WHEEL_VEL;
}
