/*
 * network.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include <main/ctrl.h>
#include <main/ctrl_calibrate.h>
#include <main/ctrl_motor.h>
#include <main/ctrl_tigga.h>
#include <main/network.h>
#include <main/skills.h>
#include "util/bits.h"
#include "util/float16_t.h"
#include "util/boot.h"
#include "util/sys_time.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/intercom_reliable.h"
#include "util/log.h"
#include "util/prog_store.h"
#include "main/prog_source_wifi.h"

#include "errors.h"
#include "commands.h"
#include "intercom_constants.h"
#include "struct_ids.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#define EVENT_REL_ACK 0x10000
#define EVENT_REL_RUN 0x20000

#define PROCESSED 0
#define PASS_ON 1

Network network = {
	.config = { 0xFF, 100, 2 },
	.rwDelay = { 300, 300, 500, 300, 1500, 500 },
};

static const ConfigFileDesc configFileDescRwDelay =
	{ SID_CFG_NET_RW_DELAY, 0, "net/rwDelay", 6, (ElementDesc[]) {
		{ UINT32, "read_2m", "", "  2M Read" },
		{ UINT32, "write_2m", "", "  2M Write" },
		{ UINT32, "read_1m", "", "  1M Read" },
		{ UINT32, "write_1m", "", "  1M Write" },
		{ UINT32, "read_250k", "", "250k Read" },
		{ UINT32, "write_250k", "", "250k Write" },
	} };

static const ConfigFileDesc configFileDescSettings =
	{ SID_CFG_NET_SETTINGS, 2, "net/config", 3, (ElementDesc[]) {
		{ UINT8, "bot_id", "", "Bot ID" },
		{ UINT8, "channel", "", "Channel" },
		{ UINT8, "speed", "", "Speed" },
	} };

static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength);

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case SID_CFG_NET_RW_DELAY:
		{
			NRF24ExSetReadWriteDelay(&network.nrf24, &network.rwDelay);
		}
		break;
		case SID_CFG_NET_SETTINGS:
		{
			if(network.config.botId != network.queue.id)
				NRF24ExChangeId(&network.nrf24, network.config.botId);

			if(network.config.channel != network.nrf24.channel)
				NRF24ExChangeChannel(&network.nrf24, network.config.channel);

			if(network.config.speed != NRF24ExGetSpeedId(&network.nrf24))
				NRF24ExChangeSpeed(&network.nrf24, network.config.speed);

			CtrlUpdateBotId(network.config.botId);
		}
		break;
	}
}

void NetworkInit()
{
	// data setup
	chBSemObjectInit(&network.timerSem, TRUE);
	chMtxObjectInit(&network.printMutex);
	chMtxObjectInit(&network.sendPacketMutex);

	FifoLinInit(&network.reliableCmds, network.reliableBuf, NETWORK_RELIABLE_BUF_SIZE);
	chMBObjectInit(&network.relEvents, network.relEventsData, NETWORK_REL_EVENT_SIZE);

	network.botDisabled = 0;

	network.pNetworkConfig = ConfigOpenOrCreate(&configFileDescSettings, &network.config, sizeof(ConfigNetwork), &configUpdateCallback);
//	network.pRwDelayConfig = ConfigOpenOrCreate(&configFileDescRwDelay, &network.rwDelay, sizeof(NRF24ReadWriteDelay), &configUpdateCallback);

	network.posCb = 0;
	network.avgRxPeriod = 0;
	network.lastUSec = 0;

	NetworkImplInit();

	NRF24ExSetReadWriteDelay(&network.nrf24, &network.rwDelay);

//	ConfigNotifyUpdate(network.pRwDelayConfig);
	ConfigNotifyUpdate(network.pNetworkConfig);
}

void NetworkSaveConfig()
{
	ConfigNotifyUpdate(network.pNetworkConfig);
}

void NetworkTask(void* params)
{
	(void)params;

	chRegSetThreadName("Network");

	uint32_t len;
	int16_t result;
	uint32_t numDelimiters = 0;
	uint32_t i;
	network.avgRxPeriod = 99999;

	NRF24ExSetup(&network.nrf24);
	NRF24ExUseQueue(&network.nrf24, &network.queue);

	network.lastMatchCtrlTime = chVTGetSystemTime()-MS2ST(2000);

	while(1)
	{
		if(network.botDisabled)
		{
			chThdSleep(MS2ST(100));
			continue;
		}

		result = NRF24ExWaitForIRQ(&network.nrf24, MS2ST(100));
		if(result)
		{
			LogErrorC("NRF24ExWaitForIRQ", result);

			NRF24ExSetup(&network.nrf24);
			NRF24ExUseQueue(&network.nrf24, &network.queue);
		}

		const uint32_t alpha = 96;
		uint32_t tNow = SysTimeUSec();
		uint32_t tPeriod = tNow - network.lastUSec;
		network.lastUSec = tNow;
		network.avgRxPeriod = (alpha*network.avgRxPeriod + (100-alpha)*tPeriod)/100;

		if(network.avgRxPeriod > 80000 && network.bsOnline)
		{
			network.bsOnline = 0;

			CtrlSetMotorsOn(0);
		}

		if(network.avgRxPeriod < 80000 && !network.bsOnline)
		{
			network.bsOnline = 1;
		}

		uint32_t readDelay;
		uint32_t writeDelay;

		NRF24ExGetReadWriteDelay(&network.nrf24, &readDelay, &writeDelay);

		chBSemReset(&network.timerSem, TRUE);
		NetworkImplStartTimeout(readDelay);
		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != MSG_OK)
		{
			LogError("Timeout");
		}

		result = NRF24ExRxDownload(&network.nrf24);
		if(result)
			LogErrorC("NRF24ExRxDownload result: 0x%02X\r\n", result);

		chBSemReset(&network.timerSem, TRUE);
		NetworkImplStartTimeout(writeDelay);

		numDelimiters = network.queue.rxFifo.numPackets;
		for(i = 0; i < numDelimiters; i++)
		{
			result = NRF24QueueGetPacket(&network.queue, network.rxProcessingBuf, NETWORK_RX_PROCESSING_SIZE, &len);
			if(result)
			{
				LogErrorC("GetPacket", result);
				break;
			}

			if(len > PACKET_HEADER_SIZE)
			{
				PacketHeader* pHeader = (PacketHeader*)network.rxProcessingBuf;
				uint8_t* pBuf = network.rxProcessingBuf+PACKET_HEADER_SIZE;
				uint16_t dataLength = len-PACKET_HEADER_SIZE;

				// Handle extended packets
				if(pHeader->section & SECTION_EXTENDED_PACKET)
				{
					PacketHeaderEx* pHeaderEx = (PacketHeaderEx*)pHeader;

					PacketHeader ack;
					ack.cmd = CMD_SYSTEM_ACK;
					ack.section = SECTION_SYSTEM;

					NetworkSendPacket(&ack, &pHeaderEx->seq, sizeof(uint16_t));

					uint8_t* pBufOrig = pBuf;
					pHeader->section &= ~SECTION_EXTENDED_PACKET;
					pBuf += PACKET_HEADER_EX_SIZE-PACKET_HEADER_SIZE;
					dataLength -= PACKET_HEADER_EX_SIZE-PACKET_HEADER_SIZE;

					memcpy(pBufOrig, pHeader, sizeof(PacketHeader));
					pHeader = (PacketHeader*)pBufOrig;
				}

				if(processPacket(pHeader, pBuf, dataLength) == PASS_ON)
				{
					NetworkImplProcessPacket(pHeader, pBuf, dataLength);
				}
			}
		}

		if(chVTTimeElapsedSinceX(network.lastUIWifiStatTime) > MS2ST(50))
		{
			network.lastUIWifiStatTime = chVTGetSystemTimeX();

			PresenterWifiStat stat;
			stat.bsOnline = network.bsOnline;
			stat.sumatraOnline = network.sumatraOnline;
			stat.updateFreq = 1000000/network.avgRxPeriod;
			stat.visionDelay = network.lastVisionDelay;

			NetworkImplWifiStatsUpdate(&stat);
		}

		if(chVTTimeElapsedSinceX(network.lastMatchCtrlTime) > MS2ST(1000))
		{
			if(network.sumatraOnline)
			{
				CtrlSetMotorsOn(0);
				SkillsSetInput(0, 0);
				network.sumatraOnline = 0;

				CtrlImplBuzzerPlay(CTRL_BUZZ_DOWN50);
			}
		}
		else
		{
			if(network.sumatraOnline == 0)
			{
				network.sumatraOnline = 1;

				CtrlImplBuzzerPlay(CTRL_BUZZ_UP50);
			}
		}

		if(chBSemWaitTimeout(&network.timerSem, MS2ST(100)) != MSG_OK)
			LogError("RxUpDelay error\r\n");

		result = NRF24ExRxUpload(&network.nrf24);
		if(result)
			LogErrorC("NRF24ExRxUpload result: 0x%02X\r\n", result);
	}
}

void NetworkReliableTask(void* params)
{
	(void)params;
	msg_t event;
	msg_t chResult;

	uint8_t* pBuf = 0;
	PacketHeaderEx* pHeaderEx = 0;
	uint32_t cmdSize;
	uint32_t lastSendTime = chVTGetSystemTime()-MS2ST(200);

	chRegSetThreadName("NetRel");

	while(1)
	{
		chResult = chMBFetch(&network.relEvents, &event, MS2ST(10));

		if(FifoLinGet(&network.reliableCmds, &pBuf, &cmdSize) == 0)
			pHeaderEx = (PacketHeaderEx*)pBuf;
		else
			pHeaderEx = 0;

		if(chResult == MSG_OK && (event & EVENT_REL_ACK) && pHeaderEx)
		{
			if((event & 0xFFFF) == pHeaderEx->seq)
			{
				FifoLinDelete(&network.reliableCmds);

				if(FifoLinGet(&network.reliableCmds, &pBuf, &cmdSize) == 0)
					pHeaderEx = (PacketHeaderEx*)pBuf;
				else
					pHeaderEx = 0;

				lastSendTime = chVTGetSystemTime()-MS2ST(200);
			}
		}

		if(pHeaderEx && chVTTimeElapsedSinceX(lastSendTime) > MS2ST(100))
		{
			chMtxLock(&network.sendPacketMutex);	// prevent re-entering

			// data to send is now in network.txProcessingBuf
			int16_t result = NRF24QueueSend(&network.queue, pBuf, cmdSize);
			if(result)
			{
				FifoLinClear(&network.queue.txFifo);
			}

			chMtxUnlock(&network.sendPacketMutex);

			lastSendTime = chVTGetSystemTime();
		}
	}
}

void NetworkSetPosCallback(PosCallback cb)
{
	network.posCb = cb;
}

int16_t NetworkSendPacket(PacketHeader* pHeader, void* _pData, uint16_t dataLength)
{
	uint8_t* pData = (uint8_t*)_pData;

	if(dataLength + PACKET_HEADER_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	memcpy(network.txProcessingBuf, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(network.txProcessingBuf + PACKET_HEADER_SIZE, pData, dataLength);

	// data to send is now in network.txProcessingBuf
	int16_t result = NRF24QueueSend(&network.queue, network.txProcessingBuf, dataLength+PACKET_HEADER_SIZE);
	if(result)
	{
		LogErrorC("NRF24QueueSend error", result);
	}
	if(result)
	{
		FifoLinClear(&network.queue.txFifo);
		LogWarn("txFifo full => cleared\r\n");
	}

	chMtxUnlock(&network.sendPacketMutex);

	return result;
}

void NetworkSendPacketRaw(void* _pData, uint16_t dataLength)
{
	uint8_t* pData = (uint8_t*)_pData;

	chMtxLock(&network.sendPacketMutex);	// prevent re-entering

	// data to send is now in network.txProcessingBuf
	int16_t result = NRF24QueueSend(&network.queue, pData, dataLength);
	if(result)
	{
		LogErrorC("NRF24QueueSend error", result);
	}
	if(result)
	{
		FifoLinClear(&network.queue.txFifo);
		LogWarn("txFifo full => cleared\r\n");
	}

	chMtxUnlock(&network.sendPacketMutex);
}

int16_t NetworkSendPacketReliable(PacketHeader* pHeader, void* _pData, uint16_t dataLength)
{
	int16_t result;
	uint8_t* pBuf;
	uint8_t* pData = (uint8_t*)_pData;

	result = FifoLinReserve(&network.reliableCmds, dataLength+PACKET_HEADER_EX_SIZE, &pBuf);
	if(result)
		return result;

	PacketHeaderEx* pHeaderEx = (PacketHeaderEx*)pBuf;
	pHeaderEx->cmd = pHeader->cmd;
	pHeaderEx->section = pHeader->section | SECTION_EXTENDED_PACKET;
	pHeaderEx->seq = network.seq++;

	memcpy(pBuf+PACKET_HEADER_EX_SIZE, pData, dataLength);

	result = FifoLinCommit(&network.reliableCmds, dataLength+PACKET_HEADER_EX_SIZE);
	if(result)
		return result;

	chMBPost(&network.relEvents, EVENT_REL_RUN, TIME_IMMEDIATE);

	return 0;
}

// network pre-processing function
// returns 0 if packet has been fully handled and should not be passed on to implementation
static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength)
{
	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_PING:
				{
					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;

					NetworkSendPacket(&header, pBuf, dataLength);

					return PROCESSED;
				}
				break;
				case CMD_SYSTEM_ACK:
				{
					uint16_t* pSeq = (uint16_t*)pBuf;

					chMBPost(&network.relEvents, EVENT_REL_ACK + *pSeq, TIME_IMMEDIATE);

					return PROCESSED;
				}
				break;
				case CMD_SYSTEM_MATCH_CTRL:
				{
					SystemMatchCtrl* pCtrl = (SystemMatchCtrl*)pBuf;

					network.lastMatchCtrlTime = chVTGetSystemTimeX();

					CtrlSetLogRate(pCtrl->feedbackFreq);

					if(pCtrl->curPosition[0] != SYSTEM_MATCH_CTRL_UNUSED_FIELD)
					{
						uint32_t delay = pCtrl->posDelay;
						network.lastVisionDelay = (delay*25000)/100;	// delay now in [us]
						network.lastVisionPosTimestamp = SysTimeUSec();
						network.lastVisionCamId = pCtrl->camId;
						memcpy(network.lastVisionPos, pCtrl->curPosition, sizeof(int16_t)*3);

						if(network.posCb)
							(*network.posCb)();
					}

					if(network.skillOverride == 0)
						SkillsSetInput(pCtrl->skillId, pCtrl->skillData);

					// Kicker/Dribbler and cheering must be handled by user
					return PASS_ON;
				}
				break;
				case CMD_SYSTEM_TIMESYNC:
				{
					if(dataLength < sizeof(uint32_t))
					{
						LogErrorC("timesync too short", dataLength);
						return PROCESSED;
					}

					uint32_t* pUnixTime = (uint32_t*)pBuf;

					sysTime.unixOffset = *pUnixTime;

					return PROCESSED;
				}
				break;
				default:
					return PASS_ON;
			}
		}
		break;
		case SECTION_BOOTLOADER:
		{
			if(pHeader->cmd == CMD_BOOTLOADER_CHECK)
			{
				ProgStoreStart(PROG_STORE_SOURCE_WIFI);
			}
			else
			{
				ProgSourceWifiNetworkInput(pHeader, dataLength, pBuf);
			}

			return PROCESSED;
		}
		break;
		case SECTION_CTRL:
		{
			switch(pHeader->cmd)
			{
				case CMD_CTRL_SET_CONTROLLER_TYPE:
				{
					CtrlSetControllerType* pCtrl = (CtrlSetControllerType*)pBuf;

					switch(pCtrl->controllerId)
					{
						case CTRL_TYPE_CALIBRATE:
						{
							CtrlSetController(&ctrlCalibrateInstance);
						}
						break;
						case CTRL_TYPE_MOTOR:
						{
							CtrlSetController(&ctrlMotorInstance);
						}
						break;
						case CTRL_TYPE_FUSION:
						{
						}
						break;
						case CTRL_TYPE_TIGGA:
						{
							CtrlSetController(&ctrlTiggaInstance);
						}
						break;
						default:
						{
							CtrlSetController(0);
						}
						break;
					}
				}
				break;
				default:
					return PASS_ON;
			}
		}
		break;
		case SECTION_CONFIG:
		{
			return PASS_ON;
		}
		break;
		default:
			return PASS_ON;
	}

	return 0;
}
