/*
 * spi_sync.c
 *
 *  Created on: 28.09.2013
 *      Author: AndreR
 */

#ifndef STM32F30X

#include "spi_sync.h"
#include "errors.h"
#include "util/bits.h"
#include "util/init_hal.h"
#include "util/log.h"

#define DMAClearEn(pDMA)	BITS_CLEAR(pDMA->CR, DMA_SxCR_EN)
#define DMASetEn(pDMA)		BITS_SET(pDMA->CR, DMA_SxCR_EN)

static void stopTransfer(SPISync* pSPI)
{
	DMAClearEn(pSPI->pDmaRx);
	DMAClearEn(pSPI->pDmaTx);

	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags
}

// IRQ -> no task switch can occur
void SPISyncDmaRxInterrupt(SPISync* pSPI)
{
	stopTransfer(pSPI);

	chSysLockFromISR();
	chBSemSignalI(&pSPI->transferSem);
	chSysUnlockFromISR();
}

void SPISyncInit(SPISync* pSPI, SPISyncData* pInit)
{
	pSPI->data = *pInit;

	chBSemObjectInit(&pSPI->transferSem, FALSE);
	chMtxObjectInit(&pSPI->taskMutex);

	pSPI->pDmaRx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDmaTx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->pDMARxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDMATxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->IFCRRxMask = DMAGenerateMaskAll(pSPI->data.dmaChRx);
	pSPI->IFCRTxMask = DMAGenerateMaskAll(pSPI->data.dmaChTx);

	uint32_t chTx;
	uint32_t chRx;

	// Select correct channel
	if(pInit->pRegister == SPI1) // DMA2 Channel3
	{
		chRx = 3;
		chTx = 3;
	}
	if(pInit->pRegister == SPI2 || pInit->pRegister == SPI3) // DMA1 Channel0
	{
		chRx = 0;
		chTx = 0;
	}
#ifdef STM32F7XX
	if(pInit->pRegister == SPI4) // DMA CH4 (ST0, ST1) or CH5 (ST3,ST4)
	{
		if(pSPI->data.dmaChRx == 0)
			chRx = 4;
		else
			chRx = 5;

		if(pSPI->data.dmaChTx == 1)
			chTx = 4;
		else
			chRx = 5;
	}
	if(pInit->pRegister == SPI5) // DMA2 CH2 (ST3,ST4) or CH7 (ST5,ST6)
	{
		if(pSPI->data.dmaChRx == 3)
			chRx = 2;
		else
			chRx = 7;

		if(pSPI->data.dmaChTx == 4)
			chTx = 2;
		else
			chRx = 7;
	}
	if(pInit->pRegister == SPI6) // DMA2 Channel1
	{
		chRx = 1;
		chTx = 1;
	}
#endif

	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags

	pSPI->pDmaRx->CR = (chRx << 25);
	pSPI->pDmaTx->CR = (chTx << 25);

	// configure DMA
	// Medium Priority, Memory Increment, Memory-to-peripheral
	pSPI->pDmaTx->CR |= (pInit->dmaPrio << 16) | DMA_SxCR_MINC | DMA_SxCR_DIR_0;
	pSPI->pDmaTx->FCR = DMA_SxFCR_FTH_0 | DMA_SxFCR_DMDIS | DMA_SxFCR_FEIE; // FIFO 1/2 threshold, direct mode disabled, FIFO error IRQ enabled
	pSPI->pDmaTx->NDTR = 0;	// Number of data to transfer
	pSPI->pDmaTx->PAR = (uint32_t)&pSPI->data.pRegister->DR;	// peripheral address
	pSPI->pDmaTx->M0AR = 0;	// configured on transfer

	// Medium Priority, Memory Increment, Peripheral-to-Memory, Transfer complete  interrupt enable
	pSPI->pDmaRx->CR |= (pInit->dmaPrio << 16) | DMA_SxCR_MINC | DMA_SxCR_TCIE;
	pSPI->pDmaRx->FCR = DMA_SxFCR_FTH_0 | DMA_SxFCR_DMDIS | DMA_SxFCR_FEIE; // FIFO 1/2 threshold, direct mode disabled, FIFO error IRQ enabled
	pSPI->pDmaRx->NDTR = 0;
	pSPI->pDmaRx->PAR = (uint32_t)&pSPI->data.pRegister->DR;
	pSPI->pDmaRx->M0AR = 0; // configured on transfer

	pSPI->data.pRegister->SR = 0;
	pSPI->data.pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI;
	pSPI->data.pRegister->CR2 = SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;
#ifdef STM32F7XX
	pSPI->data.pRegister->CR2 |= SPI_CR2_DS_0 | SPI_CR2_DS_1 | SPI_CR2_DS_2 | SPI_CR2_FRXTH;	// set 8 bit data size
#endif
}

void SPISyncSlaveInit(SPISync* pSPI, SPISyncSlave* pSlave)
{
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_OUTPUT;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_25MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(pSlave->pCSPort, pSlave->csPin, &gpioInit);

	GPIOSet(pSlave->pCSPort, pSlave->csPin);

	pSlave->pBus = pSPI;
}

void SPISyncSplitInitTransfer(SPISyncSlave* pSlave)
{
	SPISync* pSPI = pSlave->pBus;

	// Block multiple tasks entering this function
	chMtxLock(&pSPI->taskMutex);

	// configure SPI
	pSPI->data.pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE;
	pSPI->data.pRegister->CR1 |=  pSlave->prescaler | pSlave->cpol << 1 | pSlave->cpha;

	// Clear semaphore which indicates transfer is complete
	chBSemWaitTimeout(&pSPI->transferSem, TIME_IMMEDIATE);

	// select slave
	GPIOReset(pSlave->pCSPort, pSlave->csPin);
}

void SPISyncSplitStartTransfer(SPISyncSlave* pSlave, const uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	SPISync* pSPI = pSlave->pBus;

	// configure DMA
	if(pSrc == 0)
	{
		pSPI->pDmaTx->CR &= ~DMA_SxCR_MINC;
		pSPI->pDmaTx->M0AR = (uint32_t)&pSPI->dummy;
	}
	else
	{
		pSPI->pDmaTx->CR |= DMA_SxCR_MINC;
		pSPI->pDmaTx->M0AR = (uint32_t)pSrc;
	}

	if(pDst == 0)
	{
		pSPI->pDmaRx->CR &= ~DMA_SxCR_MINC;
		pSPI->pDmaRx->M0AR = (uint32_t)&pSPI->dummy;
	}
	else
	{
		pSPI->pDmaRx->CR |= DMA_SxCR_MINC;
		pSPI->pDmaRx->M0AR = (uint32_t)pDst;
	}

	pSPI->pDmaTx->NDTR = size;
	pSPI->pDmaRx->NDTR = size;

	// enable DMA
	DMASetEn(pSPI->pDmaRx);
	DMASetEn(pSPI->pDmaTx);

//	pSPI->data.pRegister->CR2 |= SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;
}

int16_t SPISyncSplitWait(SPISyncSlave* pSlave)
{
	SPISync* pSPI = pSlave->pBus;

	// check if wait is called from the task which called start transfer
	if(pSPI->taskMutex.m_owner != chThdGetSelfX())
		return ERROR_SPI_SYNC_INVALID_TASK;

	// wait for transfer complete semaphore
	msg_t result = chBSemWaitTimeout(&pSPI->transferSem, pSlave->timeoutTicks);

	if(result != MSG_OK)
	{
		stopTransfer(pSPI);

		LogError("SPISyncTimeout");

		return ERROR_SPI_SYNC_TIMEOUT;
	}

	return 0;
}

int16_t SPISyncSplitTransfer(SPISyncSlave* pSlave, const uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	SPISyncSplitStartTransfer(pSlave, pSrc, pDst, size);

	return SPISyncSplitWait(pSlave);
}

int16_t SPISyncSplitFinish(SPISyncSlave* pSlave)
{
	SPISync* pSPI = pSlave->pBus;

	// check if wait is called from the task which called start transfer
	if(pSPI->taskMutex.m_owner != chThdGetSelfX())
		return ERROR_SPI_SYNC_INVALID_TASK;

	GPIOSet(pSlave->pCSPort, pSlave->csPin);

	pSPI->data.pRegister->CR1 = 0;	// SPE disabled

#if defined(STM32F3XX) || defined(STM32F7XX)
	// flush RXFIFO of SPI
	while((pSPI->data.pRegister->SR & SPI_SR_FRLVL) != 0)
		pSPI->data.pRegister->DR;
#endif

	// this task is done
	chMtxUnlock(&pSPI->taskMutex);

	return 0;
}

void SPISyncStartTransfer(SPISyncSlave* pSlave, uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	SPISyncSplitInitTransfer(pSlave);
	SPISyncSplitStartTransfer(pSlave, pSrc, pDst, size);
}

int16_t SPISyncWait(SPISyncSlave* pSlave)
{
	int16_t result = SPISyncSplitWait(pSlave);
	SPISyncSplitFinish(pSlave);

	return result;
}

int16_t SPISyncTransfer(SPISyncSlave* pSlave, uint8_t* pSrc, uint8_t* pDst, uint32_t size)
{
	SPISyncStartTransfer(pSlave, pSrc, pDst, size);

	return SPISyncWait(pSlave);
}

#endif
